import asyncio
import json
import random
from pathlib import Path

from ollama import Message, ResponseError
from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.events import Click
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import (
    LoadingIndicator,
    Markdown,
    Static,
    TabbedContent,
)

from oterm.app.chat_edit import ChatEdit
from oterm.app.chat_rename import ChatRename
from oterm.app.mcp_prompt import MCPPrompt
from oterm.app.prompt_history import PromptHistory
from oterm.app.widgets.image import ImageAdded
from oterm.app.widgets.prompt import FlexibleInput
from oterm.ollamaclient import OllamaLLM, Options
from oterm.store.store import Store
from oterm.tools import available_tool_calls
from oterm.types import ChatModel, MessageModel
from oterm.utils import parse_response


class ChatContainer(Widget):
    ollama = OllamaLLM()
    messages: reactive[list[MessageModel]] = reactive([])
    images: list[tuple[Path, str]] = []
    BINDINGS = [
        Binding("up", "history", "history"),
        Binding(
            "escape", "cancel_inference", "cancel inference", show=False, priority=True
        ),
    ]

    def __init__(
        self,
        *children: Widget,
        messages: list[MessageModel] = [],
        chat_model: ChatModel,
        **kwargs,
    ) -> None:
        super().__init__(*children, **kwargs)

        self.messages = messages
        self.chat_model = chat_model
        history = []
        # This is wrong, the images should be a list of Image objects
        # See https://github.com/ollama/ollama-python/issues/375
        # Temp fix is to do msg.images = images  # type: ignore
        for msg_model in messages:
            message_text = msg_model.text
            msg = Message(
                role=msg_model.role,
                content=(
                    message_text
                    if msg_model.role == "user"
                    else parse_response(message_text).response
                ),
            )
            msg.images = msg_model.images  # type: ignore
            history.append(msg)

        used_tool_defs = [
            tool_def
            for tool_def in available_tool_calls()
            if tool_def["tool"] in chat_model.tools
        ]

        self.ollama = OllamaLLM(
            model=chat_model.model,
            system=chat_model.system,
            format=chat_model.format,
            options=chat_model.parameters,
            keep_alive=chat_model.keep_alive,
            history=history,
            tool_defs=used_tool_defs,
            thinking=chat_model.thinking,
        )
        self.loaded = False
        self.loading = False
        self.images = []

    def on_mount(self) -> None:
        self.query_one("#prompt").focus()

    async def load_messages(self) -> None:
        if self.loaded or self.loading:
            return
        self.loading = True
        message_container = self.query_one("#messageContainer")
        for message in self.messages:
            chat_item = ChatItem()
            chat_item.text = (
                message.text
                if message.role == "user"
                else parse_response(message.text).formatted_output
            )
            chat_item.author = message.role
            await message_container.mount(chat_item)
        message_container.scroll_end()
        self.loading = False
        self.loaded = True

    async def response_task(self, message: str) -> None:
        message_container = self.query_one("#messageContainer")

        user_chat_item = ChatItem()
        user_chat_item.text = message
        user_chat_item.author = "user"
        message_container.mount(user_chat_item)

        response_chat_item = ChatItem()
        response_chat_item.author = "assistant"
        message_container.mount(response_chat_item)
        loading = LoadingIndicator()
        await message_container.mount(loading)
        message_container.scroll_end()

        try:
            response = ""

            async for thought, text in self.ollama.stream(
                message, [img for _, img in self.images]
            ):
                response = f"<think>{thought}</think>{text}"
                response_chat_item.text = parse_response(response).formatted_output

            parsed = parse_response(response)

            # To not exhaust the tokens, remove the thought process from the history (it seems to be the common practice)
            self.ollama.history[-1].content = parsed.response  # type: ignore
            response_chat_item.text = parsed.formatted_output

            if message_container.can_view_partial(response_chat_item):
                message_container.scroll_end()

            store = await Store.get_store()

            # Create and save user message model
            user_message = MessageModel(
                id=None,
                chat_id=self.chat_model.id,  # type: ignore
                role="user",
                text=message,
                images=[img for _, img in self.images],
            )
            id = await store.save_message(user_message)
            user_message.id = id
            self.messages.append(user_message)

            # Create and save assistant message model
            assistant_message = MessageModel(
                id=None,
                chat_id=self.chat_model.id,  # type: ignore
                role="assistant",
                text=parsed.response,
                images=[],
            )
            id = await store.save_message(assistant_message)
            assistant_message.id = id
            self.messages.append(assistant_message)
            self.images = []

        except asyncio.CancelledError:
            user_chat_item.remove()
            response_chat_item.remove()
            input = self.query_one("#prompt", FlexibleInput)
            input.text = message
        except ResponseError as e:
            user_chat_item.remove()
            response_chat_item.remove()
            self.app.notify(
                f"There was an error running your request: {e}", severity="error"
            )
            message_container.scroll_end()

        finally:
            loading.remove()

    @on(FlexibleInput.Submitted)
    async def on_submit(self, event: FlexibleInput.Submitted) -> None:
        message = event.value
        input = event.input

        input.clear()
        if not message.strip():
            input.focus()
            return

        self.inference_task = asyncio.create_task(self.response_task(message))

    def key_escape(self) -> None:
        if hasattr(self, "inference_task"):
            self.inference_task.cancel()

    @work
    async def action_edit_chat(self) -> None:
        screen = ChatEdit(chat_model=self.chat_model, edit_mode=True)

        model_info = await self.app.push_screen_wait(screen)
        if model_info is None:
            return

        self.chat_model = ChatModel.model_validate_json(model_info)

        # Save to database
        store = await Store.get_store()
        await store.edit_chat(self.chat_model)

        # load the history from messages
        history: list[Message] = []
        # This is wrong, the images should be a list of Image objects
        # See https://github.com/ollama/ollama-python/issues/375
        # Temp fix is to do msg.images = images  # type: ignore
        for message in self.messages:
            msg = Message(
                role=message.role,
                content=message.text,
            )
            msg.images = message.images  # type: ignore
            history.append(msg)

        # Get tool definitions based on the updated tools list
        used_tool_defs = [
            tool_def
            for tool_def in available_tool_calls()
            if tool_def["tool"] in self.chat_model.tools
        ]

        # Recreate the Ollama client with updated parameters
        self.ollama = OllamaLLM(
            model=self.chat_model.model,
            system=self.chat_model.system,
            format=self.chat_model.format,
            options=self.chat_model.parameters,
            keep_alive=self.chat_model.keep_alive,
            history=history,  # type: ignore
            tool_defs=used_tool_defs,
            thinking=self.chat_model.thinking,
        )

    @work
    async def action_rename_chat(self) -> None:
        store = await Store.get_store()
        screen = ChatRename(self.chat_model.name)
        new_name = await self.app.push_screen_wait(screen)
        if new_name is None:
            return
        tabs = self.app.query_one(TabbedContent)
        await store.rename_chat(self.chat_model.id, new_name)  # type: ignore
        tabs.get_tab(f"chat-{self.chat_model.id}").update(new_name)
        self.app.notify("Chat renamed")

    async def action_clear_chat(self) -> None:
        self.messages = []
        self.images = []
        self.ollama = OllamaLLM(
            model=self.ollama.model,
            system=self.ollama.system,
            format=self.ollama.format,  # type: ignore
            options=self.chat_model.parameters,
            keep_alive=self.ollama.keep_alive,
            history=[],  # type: ignore
            tool_defs=self.ollama.tool_defs,
            thinking=self.chat_model.thinking,
        )
        msg_container = self.query_one("#messageContainer")
        for child in msg_container.children:
            child.remove()
        store = await Store.get_store()
        await store.clear_chat(self.chat_model.id)  # type: ignore

    async def action_regenerate_llm_message(self) -> None:
        if not self.messages[-1:]:
            return
        # Remove last Ollama response from UI and regenerate it
        response_message_id = self.messages[-1].id
        self.messages.pop()
        message_container = self.query_one("#messageContainer")
        message_container.children[-1].remove()
        response_chat_item = ChatItem()
        response_chat_item.author = "assistant"
        message_container.mount(response_chat_item)
        loading = LoadingIndicator()
        await message_container.mount(loading)
        message_container.scroll_end()

        # Remove the last two messages from chat history, we will regenerate them
        self.ollama.history = self.ollama.history[:-2]
        message = self.messages[-1]

        async def response_task() -> None:
            response = await self.ollama.completion(
                message.text,
                images=message.images,  # type: ignore
                additional_options=Options(seed=random.randint(0, 32768)),
            )
            response_chat_item.text = response
            if message_container.can_view_partial(response_chat_item):
                message_container.scroll_end()

            # Save to db
            store = await Store.get_store()

            # Create a message model for regenerated response
            regenerated_message = MessageModel(
                id=response_message_id,
                chat_id=self.chat_model.id,  # type: ignore
                role="assistant",
                text=response,
                images=[],
            )
            await store.save_message(regenerated_message)
            regenerated_message.id = response_message_id
            self.messages.append(regenerated_message)
            self.images = []
            loading.remove()

        asyncio.create_task(response_task())

    async def action_history(self) -> None:
        def on_history_selected(text: str | None) -> None:
            if text is None:
                return
            prompt = self.query_one("#prompt", FlexibleInput)
            if "\n" in text and not prompt.is_multiline:
                prompt.toggle_multiline()
            prompt.text = text
            prompt.focus()

        prompts = [message.text for message in self.messages if message.role == "user"]
        prompts.reverse()
        screen = PromptHistory(prompts)
        self.app.push_screen(screen, on_history_selected)

    @work
    async def action_mcp_prompt(self) -> None:
        screen = MCPPrompt()
        messages = await self.app.push_screen_wait(screen)
        if messages is None:
            return
        messages = [Message(**msg) for msg in json.loads(messages)]
        message_container = self.query_one("#messageContainer")
        store = await Store.get_store()

        last_user_message = None
        if messages[-1].role == "user":
            last_user_message = messages.pop()

        for message in messages:
            text = message.content or ""
            # Create a message model for the MCP conversation
            message_model = MessageModel(
                id=None,
                chat_id=self.chat_model.id,  # type: ignore
                role=message.role,  # type: ignore
                text=text,
                images=[],
            )
            id = await store.save_message(message_model)
            message_model.id = id
            self.messages.append(message_model)
            chat_item = ChatItem()
            chat_item.text = text
            chat_item.author = message.role
            await message_container.mount(chat_item)
        message_container.scroll_end()
        if last_user_message is not None and last_user_message.content:
            await self.response_task(last_user_message.content)

    @on(ImageAdded)
    def on_image_added(self, ev: ImageAdded) -> None:
        self.images.append((ev.path, ev.image))
        self.app.notify(f"Image {ev.path} added.")

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Static(f"model: {self.ollama.model}", id="info")
            yield VerticalScroll(
                id="messageContainer",
            )
            yield FlexibleInput("", id="prompt", classes="singleline")


class ChatItem(Widget):
    text: reactive[str] = reactive("")
    author: reactive[str] = reactive("")

    @on(Click)
    async def on_click(self, event: Click) -> None:
        self.app.copy_to_clipboard(self.text)
        widgets = self.query(".text")
        for widget in widgets:
            widget.styles.animate("opacity", 0.5, duration=0.1)
            widget.styles.animate("opacity", 1.0, duration=0.1, delay=0.1)
        self.app.notify("Message copied to clipboard.")

    async def watch_text(self, text: str) -> None:
        if self.author == "user":
            return
        try:
            jsn = json.loads(text)
            if isinstance(jsn, dict):
                text = f"```json\n{self.text}\n```"
        except json.JSONDecodeError:
            pass

        txt_widget = self.query_one(".text", Markdown)
        await txt_widget.update(text)

    def compose(self) -> ComposeResult:
        """A chat item."""
        mrk_down = Markdown(self.text, classes="text")
        mrk_down.code_dark_theme = "solarized-dark"
        mrk_down.code_light_theme = "solarized-light"
        with Horizontal(classes=f"{self.author} chatItem"):
            if self.author == "user":
                yield Static(self.text, classes="text")
            else:
                yield mrk_down



import json
from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite

from oterm.ollamaclient import OllamaLLM, parse_ollama_parameters


async def parameters(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                ALTER TABLE chat ADD COLUMN parameters TEXT DEFAULT "{}";
                """
            )
        except aiosqlite.OperationalError:
            pass

        # Update with default parameters
        chat_models = await connection.execute_fetchall("SELECT id, model FROM chat")
        for chat_id, model in chat_models:
            info = OllamaLLM.show(model)
            parameters = parse_ollama_parameters(info["parameters"])
            await connection.execute(
                "UPDATE chat SET parameters = ? WHERE id = ?",
                (json.dumps(parameters), chat_id),
            )
        await connection.commit()


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.3.0", [parameters])
]



import textual.widgets._markdown as markdown
from textual import on
from textual.events import Click


class MarkdownFence(markdown.MarkdownFence):
    @on(Click)
    async def on_click(self, event: Click) -> None:
        event.stop()
        self.app.copy_to_clipboard(self.code)
        self.styles.animate("opacity", 0.5, duration=0.1)
        self.styles.animate("opacity", 1.0, duration=0.1, delay=0.1)


markdown.MarkdownFence = MarkdownFence



from ollama import Options, ShowResponse
from pydantic import ValidationError
from rich.text import Text
from textual.app import ComposeResult
from textual.containers import (
    Container,
    Horizontal,
    Vertical,
)
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Input, Label, OptionList, TextArea

from oterm.app.widgets.tool_select import ToolSelector
from oterm.ollamaclient import (
    OllamaLLM,
    jsonify_options,
    parse_format,
    parse_ollama_parameters,
)
from oterm.types import ChatModel, OtermOllamaOptions, Tool


class ChatEdit(ModalScreen[str]):
    models = []
    models_info: dict[str, ShowResponse] = {}

    model_name: reactive[str] = reactive("")
    tag: reactive[str] = reactive("")
    bytes: reactive[int] = reactive(0)
    model_info: ShowResponse
    system: reactive[str] = reactive("")
    parameters: reactive[Options] = reactive(Options())
    format: reactive[str] = reactive("")
    keep_alive: reactive[int] = reactive(5)
    last_highlighted_index = None
    tools: reactive[list[Tool]] = reactive([])
    edit_mode: reactive[bool] = reactive(False)
    thinking: reactive[bool] = reactive(False)

    BINDINGS = [
        ("escape", "cancel", "Cancel"),
        ("enter", "save", "Save"),
    ]

    def __init__(
        self,
        chat_model: ChatModel | None = None,
        edit_mode: bool = False,
    ) -> None:
        super().__init__()

        if chat_model is None:
            chat_model = ChatModel()

        self.chat_model = chat_model
        self.model_name, self.tag = (
            chat_model.model.split(":") if chat_model.model else ("", "")
        )
        self.system = chat_model.system or ""
        self.parameters = chat_model.parameters
        self.format = chat_model.format
        self.keep_alive = chat_model.keep_alive
        self.tools = chat_model.tools
        self.edit_mode = edit_mode
        self.thinking = chat_model.thinking

    def _return_chat_meta(self) -> None:
        model = f"{self.model_name}:{self.tag}"
        system = self.query_one(".system", TextArea).text
        system = system if system != self.model_info.get("system", "") else None
        keep_alive = int(self.query_one(".keep-alive", Input).value)
        p_area = self.query_one(".parameters", TextArea)
        try:
            parameters = OtermOllamaOptions.model_validate_json(
                p_area.text, strict=True
            )

            if isinstance(parameters.stop, str):
                parameters.stop = [parameters.stop]

        except ValidationError:
            self.app.notify("Error validating parameters", severity="error")
            p_area = self.query_one(".parameters", TextArea)
            p_area.styles.animate("opacity", 0.0, final_value=1.0, duration=0.5)
            return

        f_area = self.query_one(".format", TextArea)
        try:
            parse_format(f_area.text)
            format = f_area.text
        except Exception:
            self.app.notify("Error parsing format", severity="error")
            f_area.styles.animate("opacity", 0.0, final_value=1.0, duration=0.5)
            return

        self.tools = self.query_one(ToolSelector).selected
        self.thinking = self.query_one("#thinking-checkbox", Checkbox).value

        # Create updated chat model
        updated_chat_model = ChatModel(
            id=self.chat_model.id,
            name=self.chat_model.name,
            model=model,
            system=system,
            format=format,
            parameters=parameters,
            keep_alive=keep_alive,
            tools=self.tools,
            thinking=self.thinking,
        )

        self.dismiss(updated_chat_model.model_dump_json(exclude_none=True))

    def action_cancel(self) -> None:
        self.dismiss()

    def action_save(self) -> None:
        self._return_chat_meta()

    def select_model(self, model: str) -> None:
        select = self.query_one("#model-select", OptionList)
        for index, option in enumerate(select._options):
            if str(option.prompt) == model:
                select.highlighted = index
                break

    async def on_mount(self) -> None:
        self.models = OllamaLLM.list().models

        models = [model.model or "" for model in self.models]
        for model in models:
            info = OllamaLLM.show(model)
            self.models_info[model] = info
        option_list = self.query_one("#model-select", OptionList)
        option_list.clear_options()
        for model in models:
            option_list.add_option(option=self.model_option(model))
        option_list.highlighted = self.last_highlighted_index
        if self.model_name and self.tag:
            self.select_model(f"{self.model_name}:{self.tag}")

        # Disable the model select widget if we are in edit mode.
        widget = self.query_one("#model-select", OptionList)
        widget.disabled = self.edit_mode

    def on_option_list_option_highlighted(
        self, option: OptionList.OptionHighlighted
    ) -> None:
        model = option.option.prompt
        model_meta = next((m for m in self.models if m.model == str(model)), None)
        if model_meta:
            name, tag = (model_meta.model or "").split(":")
            self.model_name = name
            widget = self.query_one(".name", Label)
            widget.update(f"{self.model_name}")

            self.tag = tag
            widget = self.query_one(".tag", Label)
            widget.update(f"{self.tag}")

            self.bytes = model_meta["size"]
            widget = self.query_one(".size", Label)
            widget.update(f"{(self.bytes / 1.0e9):.2f} GB")

            meta = self.models_info.get(model_meta.model or "")
            self.model_info = meta  # type: ignore
            if not self.edit_mode:
                self.parameters = parse_ollama_parameters(
                    self.model_info.parameters or ""
                )
            widget = self.query_one(".parameters", TextArea)
            widget.load_text(jsonify_options(self.parameters))
            widget = self.query_one(".system", TextArea)

            # XXX Does not work as expected, there is no longer system in model_info
            widget.load_text(self.system or self.model_info.get("system", ""))

            capabilities: list[str] = self.model_info.get("capabilities", [])
            tools_supported = "tools" in capabilities
            tool_selector = self.query_one(ToolSelector)
            tool_selector.disabled = not tools_supported

            thinking_checkbox = self.query_one("#thinking-checkbox", Checkbox)
            thinking_checkbox.disabled = "thinking" not in capabilities

            if "completion" in capabilities:
                capabilities.remove("completion")  #
            if "embedding" in capabilities:
                capabilities.remove("embedding")

            caps = (
                " ".join(capabilities)
                .replace("vision", "👁️")
                .replace("tools", "🛠️")
                .replace("thinking", "🧠")
            )
            widget = self.query_one(".caps", Label)
            widget.update(caps)

        # Now that there is a model selected we can save the chat.
        save_button = self.query_one("#save-btn", Button)
        save_button.disabled = False
        ChatEdit.last_highlighted_index = option.option_index

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.name == "save":
            self._return_chat_meta()
        else:
            self.dismiss()

    @staticmethod
    def model_option(model: str) -> Text:
        return Text(model)

    def compose(self) -> ComposeResult:
        with Container(classes="screen-container full-height"):
            with Horizontal():
                with Vertical():
                    with Horizontal(id="model-info"):
                        yield Label("Model:", classes="title")
                        yield Label(f"{self.model_name}", classes="name")
                        yield Label("Tag:", classes="title")
                        yield Label(f"{self.tag}", classes="tag")
                        yield Label("Size:", classes="title")
                        yield Label(f"{self.size}", classes="size")
                        yield Label("Caps:", classes="title")
                        yield Label("", classes="caps")

                    yield OptionList(id="model-select")
                    yield Label("Tools:", classes="title")
                    yield ToolSelector(
                        id="tool-selector-container", selected=self.tools
                    )

                with Vertical():
                    yield Label("System:", classes="title")
                    yield TextArea(self.system, classes="system log")
                    yield Label("Parameters:", classes="title")
                    yield TextArea(
                        jsonify_options(self.parameters),
                        classes="parameters log",
                        language="json",
                    )
                    yield Label("Format:", classes="title")
                    yield TextArea(
                        self.format or "",
                        classes="format log",
                        language="json",
                    )

                    with Horizontal():
                        with Horizontal():
                            yield Checkbox(
                                "Thinking",
                                id="thinking-checkbox",
                                name="thinking",
                                value=self.thinking,
                            )
                            yield Label(
                                "Keep-alive (min)", classes="title keep-alive-label"
                            )
                            yield Input(
                                classes="keep-alive", value=str(self.keep_alive)
                            )

            with Horizontal(classes="button-container"):
                yield Button(
                    "Save",
                    id="save-btn",
                    name="save",
                    disabled=True,
                    variant="primary",
                )
                yield Button("Cancel", name="cancel")



import pytest
from ollama import ResponseError

from oterm.ollamaclient import OllamaLLM
from oterm.tools.date_time import DateTimeTool


@pytest.mark.asyncio
async def test_generate(default_model):
    llm = OllamaLLM(model=default_model)
    res = await llm.completion(prompt="Please add 42 and 42")
    assert "84" in res


@pytest.mark.asyncio
async def test_llm_context(default_model):
    llm = OllamaLLM(model=default_model)
    await llm.completion("I am testing oterm, a python client for Ollama.")
    # There should now be a context saved for the conversation.
    res = await llm.completion("Do you remember what I am testing?")
    assert "oterm" in res.lower()


@pytest.mark.asyncio
async def test_multi_modal_llm(llama_image):
    llm = OllamaLLM(model="llava")
    res = await llm.completion("Describe this image", images=[llama_image])
    assert "llama" in res or "animal" in res


@pytest.mark.asyncio
async def test_errors():
    llm = OllamaLLM(model="non-existent-model")
    try:
        await llm.completion("This should fail.")
    except ResponseError as e:
        assert 'model "non-existent-model" not found' in str(e)


@pytest.mark.asyncio
async def test_iterator(default_model):
    llm = OllamaLLM(model=default_model)
    response = ""
    async for _, text in llm.stream("Please add 2 and 2"):
        response = text
    assert "4" in response


@pytest.mark.asyncio
async def test_tool_streaming(default_model):
    llm = OllamaLLM(
        model=default_model,
        tool_defs=[
            {"tool": DateTimeTool, "callable": lambda: "2025-01-01"},
        ],
    )
    response = ""
    async for _, text in llm.stream(
        "What is the current date in YYYY-MM-DD format?. Use the date_time tool to answer."
    ):
        response = text
    assert "2025-01-01" in response



from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import Label, OptionList


class PromptHistory(ModalScreen[str]):
    history: list[str] = []
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def __init__(self, history=[]) -> None:
        self.history = history
        super().__init__()

    def action_cancel(self) -> None:
        self.dismiss()

    def on_mount(self) -> None:
        option_list = self.query_one("#prompt-history", OptionList)
        option_list.clear_options()
        for prompt in self.history:
            option_list.add_option(option=Text(prompt))

    def on_option_list_option_selected(self, option: OptionList.OptionSelected) -> None:
        self.dismiss(str(option.option.prompt))

    def compose(self) -> ComposeResult:
        with Container(classes="screen-container full-height"):
            yield Label("Prompt history", classes="title")
            yield OptionList(id="prompt-history")



import asyncio

from ollama import ResponseError
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label, TextArea

from oterm.ollamaclient import OllamaLLM


class PullModel(ModalScreen[str]):
    model: str = ""
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def __init__(self, model: str) -> None:
        self.model = model
        super().__init__()

    def action_cancel(self) -> None:
        self.dismiss()

    @work
    async def pull_model(self) -> None:
        log = self.query_one(".log", TextArea)
        stream = OllamaLLM.pull(self.model)
        try:
            for response in stream:
                log.text += response.model_dump_json() + "\n"
                await asyncio.sleep(0.1)
            await asyncio.sleep(1.0)
        except ResponseError as e:
            log.text += f"Error: {e}\n"
        self.app.notify("Model pulled successfully")

    @on(Input.Changed)
    async def on_model_change(self, ev: Input.Changed) -> None:
        self.model = ev.value

    @on(Button.Pressed)
    @on(Input.Submitted)
    async def on_pull(self, ev: Button.Pressed) -> None:
        self.pull_model()

    def compose(self) -> ComposeResult:
        with Container(
            id="pull-model-container", classes="screen-container full-height"
        ):
            yield Label("Pull model", classes="title")
            with Horizontal():
                yield Input(self.model)
                yield Button("Pull", variant="primary")
            yield TextArea(classes="parameters log", read_only=True)



import itertools

from mcp.types import ImageContent, PromptMessage, TextContent
from ollama import Message

from oterm.log import log
from oterm.tools.mcp.client import MCPClient
from oterm.types import PromptCall

available_prompt_defs: dict[str, list[PromptCall]] = {}


def available_prompt_calls() -> list[PromptCall]:
    """Return a list of all available prompt calls."""
    return list(itertools.chain.from_iterable(available_prompt_defs.values()))


class MCPPromptCallable:
    def __init__(self, name: str, server_name: str, client: MCPClient):
        self.name = name
        self.server_name = server_name
        self.client = client

    async def call(self, **kwargs) -> list[PromptMessage]:
        log.info(f"Calling Prompt {self.name} in {self.server_name} with {kwargs}")
        res = await self.client.call_prompt(self.name, kwargs)
        log.info(f"Prompt {self.name} returned {res}")
        return res


def mcp_prompt_to_ollama_messages(mcp_prompt: list[PromptMessage]) -> list[Message]:
    """Convert an MCP prompt to Ollama messages"""

    messages: list[Message] = []
    for m in mcp_prompt:
        if isinstance(m.content, TextContent):
            messages.append(Message(role=m.role, content=m.content.text))
        elif isinstance(m.content, ImageContent):
            messages.append(Message(role=m.role, images=[m.content.data]))  # type: ignore

    return messages



from collections.abc import Awaitable, Callable
from typing import Any, Literal, TypedDict

from mcp.types import Prompt
from ollama import Options, Tool
from pydantic import BaseModel, Field


class ParsedResponse(BaseModel):
    thought: str
    response: str
    formatted_output: str


class ToolCall(TypedDict):
    tool: Tool
    callable: Callable | Awaitable


class PromptCall(TypedDict):
    prompt: Prompt
    callable: Callable | Awaitable


class ExternalToolDefinition(TypedDict):
    tool: str
    callable: str


class OtermOllamaOptions(Options):
    # Patch stop to allow for a single string.
    # This is an issue with the gemma model which has a single stop parameter.
    # Remove when fixed upstream and close #187
    # Using 'any' to avoid type conflict with parent class
    stop: Any = None  # type: ignore

    class Config:
        extra = "forbid"


class ChatModel(BaseModel):
    """Chat model for storing chat metadata"""

    id: int | None = None
    name: str = ""
    model: str = ""
    system: str | None = None
    format: str = ""
    parameters: OtermOllamaOptions = Field(default_factory=OtermOllamaOptions)
    keep_alive: int = 5
    tools: list[Tool] = Field(default_factory=list)
    thinking: bool = False


class MessageModel(BaseModel):
    """Message model for storing chat messages"""

    id: int | None = None
    chat_id: int
    role: Literal["user", "assistant", "system", "tool"]
    text: str
    images: list[str] = Field(default_factory=list)



from collections.abc import AsyncGenerator
from io import BytesIO
from pathlib import Path

import ollama
import pytest
import pytest_asyncio
from mcp import StdioServerParameters
from PIL import Image

from oterm.tools.mcp.client import MCPClient

DEFAULT_MODEL = "llama3.2"


@pytest_asyncio.fixture(autouse=True)
async def load_test_models():
    try:
        ollama.show(DEFAULT_MODEL)
    except ollama.ResponseError:
        ollama.pull(DEFAULT_MODEL)
    yield


@pytest.fixture(scope="session")
def default_model() -> str:
    return DEFAULT_MODEL


@pytest.fixture(scope="session")
def llama_image() -> bytes:
    buffered = BytesIO()
    image = Image.open("tests/data/lama.jpg")
    image.save(buffered, format="JPEG")
    return buffered.getvalue()


@pytest.fixture(scope="session")
def mcp_server_config() -> dict:
    mcp_server_executable = Path(__file__).parent / "tools" / "mcp_servers.py"
    return {
        "stdio": {
            "command": "mcp",
            "args": ["run", mcp_server_executable.absolute().as_posix()],
        },
        "sse": {
            "url": "http://localhost:8000/sse",
        },
        "ws": {
            "url": "ws://localhost:8000/ws",
        },
    }


@pytest_asyncio.fixture(scope="function")
async def mcp_client(mcp_server_config) -> AsyncGenerator[MCPClient, None]:
    client = MCPClient(
        "test_server",
        StdioServerParameters.model_validate(mcp_server_config["stdio"]),
    )
    await client.initialize()

    yield client
    await client.teardown()



import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.think import ThinkTool, think


@pytest.mark.asyncio
async def test_think(default_model):
    llm = OllamaLLM(
        model=default_model, tool_defs=[{"tool": ThinkTool, "callable": think}]
    )
    res = await llm.completion(
        """
Cannibals ambush a safari in the jungle and capture three men. The cannibals give the men a single chance to escape uneaten.
The captives are lined up in order of height, and are tied to stakes. The man in the rear can see the backs of his two friends, the man in the middle can see the back of the man in front, and the man in front cannot see anyone. The cannibals show the men five hats. Three of the hats are black and two of the hats are white.
Blindfolds are then placed over each man's eyes and a hat is placed on each man's head. The two hats left over are hidden. The blindfolds are then removed and it is said to the men that if one of them can guess what color hat he is wearing they can all leave unharmed.
The man in the rear who can see both of his friends' hats but not his own says, "I don't know". The middle man who can see the hat of the man in front, but not his own says, "I don't know". The front man who cannot see ANYBODY'S hat says "I know!"
What was the color of his hat? Reply just with the color of the hat.
        """
    )
    assert res.lower() == "black" or res.lower() == "white"



from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def update_roles(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        await connection.executescript(
            """
            UPDATE message SET author = 'assistant' WHERE author = 'ollama';
            UPDATE message SET author = 'user' WHERE author = 'me';
            """
        )


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.12.0", [update_roles]),
]



import pytest
from mcp.types import Prompt, PromptMessage, TextContent
from ollama import Message

from oterm.tools.mcp.prompts import MCPPromptCallable, mcp_prompt_to_ollama_messages


@pytest.mark.asyncio
async def test_mcp_simple_string_prompt(mcp_client):
    await mcp_client.initialize()
    prompts = await mcp_client.get_available_prompts()
    for prompt in prompts:
        assert Prompt.model_validate(prompt)

    oracle_prompt = [p for p in prompts if p.name == "oracle_prompt"][0]
    assert oracle_prompt.name == "oracle_prompt"
    assert oracle_prompt.description == "Prompt to ask the oracle a question."
    args = oracle_prompt.arguments or []
    assert len(args) == 1
    arg = args[0]
    assert arg.name == "question"
    assert arg.required

    mcpPromptCallable = MCPPromptCallable(oracle_prompt.name, "test_server", mcp_client)
    res = await mcpPromptCallable.call(question="What is the best client for Ollama?")

    assert res == [
        PromptMessage(
            role="user",
            content=TextContent(
                type="text",
                text="Oracle: What is the best client for Ollama?",
                annotations=None,
            ),
        ),
    ]

    assert mcp_prompt_to_ollama_messages(res) == [
        Message(role="user", content="Oracle: What is the best client for Ollama?")
    ]


@pytest.mark.asyncio
async def test_mcp_multiple_messages_prompt(mcp_client):
    prompts = await mcp_client.get_available_prompts()
    for prompt in prompts:
        assert Prompt.model_validate(prompt)

    debug_prompt = [p for p in prompts if p.name == "debug_error"][0]
    assert debug_prompt.name == "debug_error"
    assert debug_prompt.description == "Prompt to debug an error."
    args = debug_prompt.arguments or []
    assert len(args) == 2
    arg = args[0]
    assert arg.name == "error"
    assert arg.required

    arg = args[1]
    assert arg.name == "language"
    assert arg.required == False  # noqa

    mcpPromptCallable = MCPPromptCallable(debug_prompt.name, "test_server", mcp_client)
    res = await mcpPromptCallable.call(error="Assertion error")

    assert res == [
        PromptMessage(
            role="user",
            content=TextContent(
                type="text",
                text="I'm seeing this python error: Assertion error",
                annotations=None,
            ),
        ),
        PromptMessage(
            role="assistant",
            content=TextContent(
                type="text",
                text="I'll help debug that. What have you tried so far?",
                annotations=None,
            ),
        ),
    ]

    assert mcp_prompt_to_ollama_messages(res) == [
        Message(role="user", content="I'm seeing this python error: Assertion error"),
        Message(
            role="assistant",
            content="I'll help debug that. What have you tried so far?",
        ),
    ]



from difflib import get_close_matches

from mcp.shared.context import RequestContext
from mcp.types import (
    CreateMessageRequestParams,
    CreateMessageResult,
    ModelHint,
    SamplingMessage,
    TextContent,
)
from ollama import ListResponse, Message, Options

from oterm.log import log
from oterm.ollamaclient import OllamaLLM

_DEFAULT_MODEL = "llama3.2"


async def sampling_handler(
    messages: list[SamplingMessage],
    params: CreateMessageRequestParams,
    context: RequestContext,
) -> CreateMessageResult:
    """Handle sampling messages.

    Args:
        context: The request context.
        params: The parameters for the message.

    Returns:
        The result of the sampling.
    """
    log.info("Request for sampling", params.model_dump_json())
    msgs = [
        Message(role=msg.role, content=msg.content.text)
        for msg in messages
        if type(msg.content) is TextContent
    ]
    system = params.systemPrompt
    options = Options(temperature=params.temperature, stop=params.stopSequences)
    model = _DEFAULT_MODEL
    if params.modelPreferences and params.modelPreferences.hints:
        model_hints = params.modelPreferences.hints
        model_from_hints = await search_model(model_hints)
        if model_from_hints:
            model = model_from_hints.model or _DEFAULT_MODEL
    client = OllamaLLM(
        model=model,
        system=system,
        history=msgs,
        options=options,
    )
    response = await client.completion()

    return CreateMessageResult(
        content=TextContent(text=response, type="text"),
        role="user",
        model=model,
    )


async def search_model(hints: list[ModelHint]) -> ListResponse.Model | None:
    """
    Fuzzy search for a model.
    """
    log.info("Searching for model based on hints", [h.name for h in hints])
    available_models = OllamaLLM.list().models
    available_model_names = [model.model for model in available_models if model.model]

    hint = " ".join([h.name for h in hints if h.name])
    matches = get_close_matches(hint, available_model_names, n=1, cutoff=0.1)
    if matches:
        # Return the first matching model
        for model in available_models:
            if model.model == matches[0]:
                log.info("Found matching model", model.model)
                return model

    # If no matches are found, return None
    log.warning("No matching model found for the provided hints.")
    return None



import subprocess

from oterm.types import Tool

ShellTool = Tool(
    type="function",
    function=Tool.Function(
        name="shell",
        description="Function to execute commands in the user's shell and return the output.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={
                "command": Tool.Function.Parameters.Property(
                    type="string", description="The shell command to execute."
                )
            },
            required=["command"],
        ),
    ),
)


def shell_command(command="") -> str:
    return subprocess.run(command, shell=True, capture_output=True).stdout.decode(
        "utf-8"
    )



from datetime import datetime

from oterm.types import Tool

DateTimeTool = Tool(
    type="function",
    function=Tool.Function(
        name="date_time",
        description="Function to get the current date and time",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={},
            required=[],
        ),
    ),
)


def date_time() -> str:
    return datetime.now().isoformat()



import inspect
import json
from ast import literal_eval
from collections.abc import AsyncGenerator, AsyncIterator, Iterator, Mapping, Sequence
from pathlib import Path
from typing import (
    Any,
    Literal,
)

from ollama import (
    AsyncClient,
    ChatResponse,
    Client,
    ListResponse,
    Message,
    Options,
    ProgressResponse,
    ShowResponse,
)
from pydantic.json_schema import JsonSchemaValue

from oterm.config import envConfig
from oterm.log import log
from oterm.types import ToolCall


def parse_format(format_text: str) -> JsonSchemaValue | Literal["", "json"]:
    try:
        jsn = json.loads(format_text)
        if isinstance(jsn, dict):
            return jsn
    except json.JSONDecodeError:
        if format_text in ("", "json"):
            return format_text
    raise Exception(f"Invalid Ollama format: '{format_text}'")


class OllamaLLM:
    def __init__(
        self,
        model="llama3.2",
        system: str | None = None,
        history: Sequence[Mapping[str, Any] | Message] = [],
        format: str = "",
        options: Options = Options(),
        keep_alive: int = 5,
        tool_defs: Sequence[ToolCall] = [],
        thinking: bool = False,
    ):
        self.model = model
        self.system = system
        self.history = list(history)
        self.format = format
        self.keep_alive = keep_alive
        self.options = options
        self.tool_defs = tool_defs
        self.tools = [tool["tool"] for tool in tool_defs]
        self.thinking = thinking
        if system:
            system_prompt: Message = Message(role="system", content=system)
            self.history = [system_prompt] + self.history

    async def completion(
        self,
        prompt: str = "",
        images: list[Path | bytes | str] = [],
        additional_options: Options = Options(),
        tool_call_messages=[],
    ) -> str:
        client = AsyncClient(
            host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL
        )
        if prompt:
            user_prompt: Message = Message(role="user", content=prompt)
            if images:
                # This is a bug in Ollama the images should be a list of Image objects
                # user_prompt.images = [Image(value=image) for image in images]
                user_prompt.images = images  # type: ignore
            self.history.append(user_prompt)
        options = {
            k: v for k, v in self.options.model_dump().items() if v is not None
        } | {k: v for k, v in additional_options.model_dump().items() if v is not None}

        response: ChatResponse = await client.chat(
            model=self.model,
            messages=self.history + tool_call_messages,
            keep_alive=f"{self.keep_alive}m",
            options=options,
            format=parse_format(self.format),
            tools=self.tools,
            think=self.thinking,
        )
        message = response.message
        tool_calls = message.tool_calls
        if tool_calls:
            tool_messages = [message]
            for tool_call in tool_calls:
                tool_name = tool_call["function"]["name"]
                for tool_def in self.tool_defs:
                    if tool_def["tool"]["function"]["name"] == tool_name:
                        tool_callable = tool_def["callable"]
                        tool_arguments = tool_call["function"]["arguments"]
                        try:
                            log.debug(
                                f"Calling tool: {tool_name} with {tool_arguments}"
                            )
                            if inspect.iscoroutinefunction(tool_callable):
                                tool_response = await tool_callable(**tool_arguments)  # type: ignore
                            else:
                                tool_response = tool_callable(**tool_arguments)  # type: ignore
                            log.debug("Tool response:", tool_response)
                        except Exception as e:
                            log.error(f"Error calling tool {tool_name}", e)
                            tool_response = str(e)
                        tool_messages.append(
                            {  # type: ignore
                                "role": "tool",
                                "content": tool_response,
                                "name": tool_name,
                            }
                        )
            return await self.completion(
                tool_call_messages=tool_messages,
            )

        self.history.append(message)
        text_response = message.content
        return text_response or ""

    async def stream(
        self,
        prompt: str = "",
        images: list[Path | bytes | str] = [],
        additional_options: Options = Options(),
        tool_call_messages: list = [],
    ) -> AsyncGenerator[tuple[str, str], Any]:
        """Stream a chat response with support for tool calls.

        When tool calls are encountered during streaming, they are executed after the stream
        completes and the result is incorporated into the response.
        """
        client = AsyncClient(
            host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL
        )

        # Add user prompt to history if provided
        if prompt:
            user_prompt: Message = Message(role="user", content=prompt)
            if images:
                user_prompt.images = images  # type: ignore
            self.history.append(user_prompt)

        # Process options
        options = {
            k: v for k, v in self.options.model_dump().items() if v is not None
        } | {k: v for k, v in additional_options.model_dump().items() if v is not None}

        # Start the streaming chat
        stream: AsyncIterator[ChatResponse] = await client.chat(
            model=self.model,
            messages=self.history + tool_call_messages,
            options=options,
            keep_alive=f"{self.keep_alive}m",
            format=parse_format(self.format),
            tools=self.tools,
            stream=True,
            think=self.thinking,
        )

        text = ""
        thought = ""
        current_message = None
        pending_tool_call = False

        async for response in stream:
            message = response.message
            content = message.content if message.content else ""
            thinking = message.thinking if message.thinking else ""
            tool_calls = message.tool_calls
            # If we have tool calls, process them at the end of the stream
            if tool_calls and not pending_tool_call:
                pending_tool_call = True
                current_message = message

            # Add content to the accumulated text only if not processing tool calls
            if not pending_tool_call and (content or thinking):
                if thinking:
                    thought += thinking
                if content:
                    text += content
                yield thought, text

        # After streaming is complete, handle any tool calls
        if pending_tool_call and current_message and current_message.tool_calls:
            # Process each tool call
            tool_messages = [current_message]  # type: ignore
            for tool_call in current_message.tool_calls:
                tool_name = tool_call["function"]["name"]

                tool_args = tool_call["function"]["arguments"]
                # Execute the tool
                for tool_def in self.tool_defs:
                    if tool_def["tool"]["function"]["name"] == tool_name:
                        tool_callable = tool_def["callable"]

                        try:
                            log.debug(f"Calling tool: {tool_name} with {tool_args}")
                            # Execute the tool
                            if inspect.iscoroutinefunction(tool_callable):
                                tool_response = await tool_callable(**tool_args)  # type: ignore
                            else:
                                tool_response = tool_callable(**tool_args)  # type: ignore

                            log.debug(f"Tool response: {tool_response}")

                            # Add tool response to messages
                            tool_messages.append(
                                {  # type: ignore
                                    "role": "tool",
                                    "content": str(tool_response),
                                    "name": tool_name,
                                }
                            )
                        except Exception as e:
                            log.error(f"Error calling tool {tool_name}: {e}")
                            return

            # Use a new variable for the follow-up response to avoid duplication
            async for thought_chunk, text_chunk in self.stream(
                tool_call_messages=tool_messages,
                additional_options=additional_options,
            ):
                yield thought_chunk, text_chunk
                # Append the final text and thought to the history
                log.debug(text, thought)
                text = text_chunk
                thought = thought_chunk

        elif text:  # Only regular content was present
            self.history.append(Message(role="assistant", content=text))

    @staticmethod
    def list() -> ListResponse:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        return client.list()

    @staticmethod
    def show(model: str) -> ShowResponse:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        return client.show(model)

    @staticmethod
    def pull(model: str) -> Iterator[ProgressResponse]:
        client = Client(host=envConfig.OLLAMA_URL, verify=envConfig.OTERM_VERIFY_SSL)
        stream: Iterator[ProgressResponse] = client.pull(model, stream=True)
        yield from stream


def parse_ollama_parameters(parameter_text: str) -> Options:
    lines = parameter_text.split("\n")
    params = Options()
    valid_params = set(Options.model_fields.keys())
    for line in lines:
        if line:
            key, value = line.split(maxsplit=1)
            try:
                value = literal_eval(value)
            except (SyntaxError, ValueError):
                pass
            if key not in valid_params:
                continue
            if params.get(key):
                if not isinstance(params[key], list):
                    params[key] = [params[key], value]
                else:
                    params[key].append(value)
            else:
                params[key] = value
    return params


def jsonify_options(options: Options) -> str:
    return json.dumps(
        {
            key: value
            for key, value in options.model_dump().items()
            if value is not None
        },
        indent=2,
    )



from textual.app import ComposeResult
from textual.containers import Container
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.widgets import Label, RichLog

from oterm.log import log_lines
from oterm.utils import debounce


class LogViewer(ModalScreen[str]):
    line_count: reactive[int] = reactive(0)

    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def action_cancel(self) -> None:
        self.dismiss()

    @debounce(0.5)
    async def log_update(self) -> None:
        widget = self.query_one(RichLog)
        new_lines = log_lines[self.line_count :]
        self.line_count += len(new_lines)
        for group, line in new_lines:
            widget.write(f"[b]{group.name}[/b] - {line}")
        await self.log_update()

    async def on_screen_resume(self) -> None:
        await self.log_update()

    def compose(self) -> ComposeResult:
        with Container(id="log-viewer", classes="screen-container full-height"):
            yield Label("oterm logs", classes="title")
            yield RichLog(
                highlight=True,
                markup=True,
                auto_scroll=True,
                wrap=True,
            )



import re
import unicodedata
from collections.abc import Sequence

from textual import on
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import Input, Label

from oterm.store.store import Store
from oterm.types import MessageModel


def slugify(value):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    """
    value = str(value)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


class ChatExport(ModalScreen[str]):
    chat_id: int
    file_name: str = ""
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def __init__(self, chat_id: int, file_name: str = "") -> None:
        super().__init__()
        self.chat_id = chat_id
        self.file_name = file_name

    def action_cancel(self) -> None:
        self.dismiss()

    @on(Input.Submitted)
    async def on_submit(self, event: Input.Submitted) -> None:
        store = await Store.get_store()

        if not event.value:
            return

        messages: Sequence[MessageModel] = await store.get_messages(self.chat_id)
        with open(event.value, "w", encoding="utf-8") as file:
            for message in messages:
                file.write(f"*{message.role}*\n")
                file.write(f"{message.text}\n")
                file.write("\n---\n")
        self.app.notify(f"Chat exported to {file.name}")
        self.dismiss()

    def compose(self) -> ComposeResult:
        with Container(classes="screen-container short"):
            yield Label("Export chat", classes="title")
            yield Input(id="chat-name-input", value=self.file_name)



import pytest
from ollama import ResponseError

from oterm.ollamaclient import OllamaLLM, jsonify_options, parse_ollama_parameters


def test_list():
    llm = OllamaLLM()
    response = llm.list()
    models = response.get("models", [])
    assert [model for model in models if model.model == "llama3.2:latest"]


def test_show():
    llm = OllamaLLM()
    response = llm.show("llama3.2:latest")
    assert response
    assert response.modelfile
    assert response.parameters
    assert response.template
    assert response.details
    assert response.modelinfo
    assert response.capabilities

    assert "tools" in response.capabilities
    assert "completion" in response.capabilities
    params = parse_ollama_parameters(response.parameters)
    assert params.stop == ["<|start_header_id|>", "<|end_header_id|>", "<|eot_id|>"]
    assert params.temperature is None
    json = jsonify_options(params)
    assert json == (
        "{\n"
        '  "stop": [\n'
        '    "<|start_header_id|>",\n'
        '    "<|end_header_id|>",\n'
        '    "<|eot_id|>"\n'
        "  ]\n"
        "}"
    )


def test_pull():
    llm = OllamaLLM()
    stream = llm.pull("llama3.2:latest")
    entries = [entry.status for entry in stream]
    assert "pulling manifest" in entries
    assert "success" in entries

    with pytest.raises(ResponseError) as excinfo:
        stream = llm.pull("non-existing:latest")
        entries = [entry for entry in stream]
        assert excinfo.value == "pull model manifest: file does not exist"
        assert "success" not in entries



import json
from importlib import metadata
from pathlib import Path

import aiosqlite
from packaging.version import parse

from oterm.config import envConfig
from oterm.store.upgrades import upgrades
from oterm.types import ChatModel, MessageModel, Tool
from oterm.utils import int_to_semantic_version, semantic_version_to_int


class Store:
    db_path: Path

    _store: "Store | None" = None

    @classmethod
    async def get_store(cls) -> "Store":
        if cls._store is not None:
            return cls._store
        self = Store()
        data_path = envConfig.OTERM_DATA_DIR
        data_path.mkdir(parents=True, exist_ok=True)
        self.db_path = data_path / "store.db"

        if not self.db_path.exists():
            # Create tables and set user_version
            async with aiosqlite.connect(self.db_path) as connection:
                await connection.executescript(
                    """
                    CREATE TABLE IF NOT EXISTS "chat" (
                        "id"            INTEGER,
                        "name"          TEXT,
                        "model"         TEXT NOT NULL,
                        "system"        TEXT,
                        "format"        TEXT,
                        "parameters"    TEXT DEFAULT "{}",
                        "keep_alive"    INTEGER DEFAULT 5,
                        "tools"         TEXT DEFAULT "[]",
                        "thinking"      BOOLEAN DEFAULT 0,
                        PRIMARY KEY("id" AUTOINCREMENT)
                    );

                    CREATE TABLE IF NOT EXISTS "message" (
                        "id"		INTEGER,
                        "chat_id"	INTEGER NOT NULL,
                        "author"	TEXT NOT NULL,
                        "text"		TEXT NOT NULL,
                        "images"    TEXT DEFAULT "[]",
                        PRIMARY KEY("id" AUTOINCREMENT)
                        FOREIGN KEY("chat_id") REFERENCES "chat"("id") ON DELETE CASCADE
                    );
                """
                )
                await self.set_user_version(metadata.version("oterm"))
        else:
            # Upgrade database
            current_version: str = metadata.version("oterm")
            db_version = await self.get_user_version()
            for version, steps in upgrades:
                if parse(current_version) >= parse(version) and parse(version) > parse(
                    db_version
                ):
                    for step in steps:
                        await step(self.db_path)
            await self.set_user_version(current_version)
        cls._store = self
        return self

    async def get_user_version(self) -> str:
        async with aiosqlite.connect(self.db_path) as connection:
            res = await connection.execute("PRAGMA user_version;")
            res = await res.fetchone()
            return int_to_semantic_version(res[0] if res else 0)

    async def set_user_version(self, version: str) -> None:
        async with aiosqlite.connect(self.db_path) as connection:
            await connection.execute(
                f"PRAGMA user_version = {semantic_version_to_int(version)};"
            )

    async def save_chat(self, chat_model: ChatModel) -> int:
        async with aiosqlite.connect(self.db_path) as connection:
            res = await connection.execute_insert(
                """
                INSERT OR REPLACE
                INTO chat(id, name, model, system, format, parameters, keep_alive, tools, thinking)
                VALUES(:id, :name, :model, :system, :format, :parameters, :keep_alive, :tools, :thinking) RETURNING id;""",
                {
                    "id": chat_model.id,
                    "name": chat_model.name,
                    "model": chat_model.model,
                    "system": chat_model.system,
                    "format": chat_model.format,
                    "parameters": json.dumps(
                        chat_model.parameters.model_dump(exclude_unset=True)
                    ),
                    "keep_alive": chat_model.keep_alive,
                    "tools": json.dumps(
                        [tool.model_dump() for tool in chat_model.tools]
                    ),
                    "thinking": chat_model.thinking,
                },
            )
            await connection.commit()

        return res[0] if res else 0

    async def rename_chat(self, id: int, name: str) -> None:
        async with aiosqlite.connect(self.db_path) as connection:
            await connection.execute(
                "UPDATE chat SET name = :name WHERE id = :id;", {"id": id, "name": name}
            )
            await connection.commit()

    async def edit_chat(self, chat_model: ChatModel) -> None:
        async with aiosqlite.connect(self.db_path) as connection:
            await connection.execute(
                """
                UPDATE chat
                SET name = :name,
                    system = :system,
                    format = :format,
                    parameters = :parameters,
                    keep_alive = :keep_alive,
                    tools = :tools,
                    thinking = :thinking
                WHERE id = :id;
                """,
                {
                    "id": chat_model.id,
                    "name": chat_model.name,
                    "system": chat_model.system,
                    "format": chat_model.format,
                    "parameters": json.dumps(
                        chat_model.parameters.model_dump(exclude_unset=True)
                    ),
                    "keep_alive": chat_model.keep_alive,
                    "tools": json.dumps(
                        [tool.model_dump() for tool in chat_model.tools]
                    ),
                    "thinking": chat_model.thinking,
                },
            )
            await connection.commit()

    async def get_chats(self) -> list[ChatModel]:
        async with aiosqlite.connect(self.db_path) as connection:
            chats = await connection.execute_fetchall(
                """
                SELECT id, name, model, system, format, parameters, keep_alive, tools, thinking
                FROM chat;
                """
            )

            return [
                ChatModel(
                    id=id,
                    name=name,
                    model=model,
                    system=system,
                    format=format,
                    parameters=json.loads(parameters),
                    keep_alive=keep_alive,
                    tools=[Tool(**t) for t in json.loads(tools)],
                    thinking=thinking,
                )
                for id, name, model, system, format, parameters, keep_alive, tools, thinking in chats
            ]

    async def get_chat(self, id: int) -> ChatModel | None:
        async with aiosqlite.connect(self.db_path) as connection:
            chat = await connection.execute_fetchall(
                """
                SELECT id, name, model, system, format, parameters, keep_alive, tools, thinking
                FROM chat
                WHERE id = :id;
                """,
                {"id": id},
            )
            chat = next(iter(chat), None)
            if chat:
                (
                    id,
                    name,
                    model,
                    system,
                    format,
                    parameters,
                    keep_alive,
                    tools,
                    thinking,
                ) = chat
                return ChatModel(
                    id=id,
                    name=name,
                    model=model,
                    system=system,
                    format=format,
                    parameters=json.loads(parameters),
                    keep_alive=keep_alive,
                    tools=[Tool(**t) for t in json.loads(tools)],
                    thinking=thinking,
                )
            return None

    async def delete_chat(self, id: int) -> None:
        async with aiosqlite.connect(self.db_path) as connection:
            await connection.execute("PRAGMA foreign_keys = on;")
            await connection.execute("DELETE FROM chat WHERE id = :id;", {"id": id})
            await connection.commit()

    async def save_message(self, message_model: MessageModel) -> int:
        async with aiosqlite.connect(self.db_path) as connection:
            res = await connection.execute_insert(
                """
                INSERT OR REPLACE
                INTO message(id, chat_id, author, text, images)
                VALUES(:id, :chat_id, :author, :text, :images) RETURNING id;
                """,
                {
                    "id": message_model.id,
                    "chat_id": message_model.chat_id,
                    "author": message_model.role,
                    "text": message_model.text,
                    "images": json.dumps(message_model.images),
                },
            )
            await connection.commit()
            return res[0] if res else 0

    async def get_messages(self, chat_id: int) -> list[MessageModel]:
        async with aiosqlite.connect(self.db_path) as connection:
            messages = await connection.execute_fetchall(
                """
                SELECT id, author, text, images
                FROM message
                WHERE chat_id = :chat_id;
                """,
                {"chat_id": chat_id},
            )
            return [
                MessageModel(
                    id=id,
                    chat_id=chat_id,
                    role=author,
                    text=text,
                    images=json.loads(images),
                )
                for id, author, text, images in messages
            ]

    async def clear_chat(self, chat_id: int) -> None:
        async with aiosqlite.connect(self.db_path) as connection:
            await connection.execute(
                "DELETE FROM message WHERE chat_id = :chat_id;", {"chat_id": chat_id}
            )
            await connection.commit()



# Example upgrade script to remove type column from chat table
from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def remove_type_column(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        await connection.executescript(
            """
            BEGIN TRANSACTION;

            CREATE TABLE chat_new (
                "id"        INTEGER,
                "name"      TEXT,
                "model"     TEXT NOT NULL,
                "system"    TEXT,
                "format"    TEXT,
                "parameters" TEXT DEFAULT "{}",
                "keep_alive" INTEGER DEFAULT 5,
                "tools"     TEXT DEFAULT "[]",
                PRIMARY KEY("id" AUTOINCREMENT)
            );

            INSERT INTO chat_new(id, name, model, system, format, parameters, keep_alive, tools)
            SELECT id, name, model, system, format, parameters, keep_alive, tools FROM chat;

            DROP TABLE chat;

            ALTER TABLE chat_new RENAME TO chat;

            COMMIT;
            """
        )


async def add_thinking_column(db_path):
    """Add thinking column to chat table."""
    async with aiosqlite.connect(db_path) as connection:
        await connection.execute(
            "ALTER TABLE chat ADD COLUMN thinking BOOLEAN DEFAULT 0"
        )
        await connection.commit()


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.13.1", [remove_type_column, add_thinking_column])
]



from ollama import Tool
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, ScrollableContainer, Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Checkbox

from oterm.tools import available_tool_calls, available_tool_defs


class ToolSelector(Widget):
    selected: reactive[list[Tool]] = reactive([])

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        selected: list[Tool] = [],
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.selected = selected if selected is not None else []
        # Check if selected tools are still available from the loaded tools.
        available = [tool_def["tool"] for tool_def in available_tool_calls()]
        self.selected = [tool for tool in selected if tool in available]

    def on_mount(self) -> None:
        pass

    @on(Checkbox.Changed)
    def on_checkbox_toggled(self, ev: Checkbox.Changed):
        name = ev.control.name
        checked = ev.value

        for server in available_tool_defs.keys():
            if server == name:
                # We don't need to change selected,
                # it will change when the checkboxes are checked/unchecked,
                for tool_def in available_tool_defs[server]:
                    tool_checkbox = self.query_one(
                        f"#{server}-{tool_def['tool']['function']['name']}", Checkbox
                    )
                    if tool_checkbox.value != checked:
                        tool_checkbox.value = checked
                return

        for tool_def in available_tool_calls():
            if tool_def["tool"].function.name == str(name):  # type: ignore
                tool = tool_def["tool"]
                if checked:
                    self.selected.append(tool)
                else:
                    try:
                        self.selected.remove(tool)
                    except ValueError:
                        pass
                break

    def all_server_tools_selected(self, server: str) -> bool:
        """Check if all tools from a server are selected."""
        for tool_def in available_tool_defs[server]:
            if tool_def["tool"] not in self.selected:
                return False
        return True

    def compose(self) -> ComposeResult:
        with ScrollableContainer(
            id="tool-selector",
        ):
            for server in available_tool_defs:
                with Horizontal(classes="tool-group"):
                    yield Checkbox(
                        name=server,
                        label=server,
                        tooltip=f"Select all tools from {server}",
                        classes="tool-group-select-all",
                        value=self.all_server_tools_selected(server),
                    )
                    with Vertical(classes="tools"):
                        for tool_def in available_tool_defs[server]:
                            name = tool_def["tool"]["function"]["name"]
                            yield Checkbox(
                                id=f"{server}-{name}",
                                name=name,
                                label=name,
                                tooltip=f"{tool_def['tool']['function']['description']}",
                                value=tool_def["tool"] in self.selected,
                                classes="tool",
                            )



import asyncio
from typing import Annotated, Any

from fastmcp.client import Client
from fastmcp.client.transports import SSETransport, StdioTransport, WSTransport
from mcp import McpError, StdioServerParameters
from mcp import Tool as MCPTool
from mcp.types import (
    EmbeddedResource,
    ImageContent,
    Prompt,
    PromptMessage,
    TextContent,
)
from pydantic import BaseModel, ValidationError

from oterm.log import log
from oterm.tools.mcp.logging import Logger
from oterm.tools.mcp.sampling import sampling_handler

IsSSEURL = Annotated[str, lambda v: v.startswith("http://") or v.startswith("https://")]


class SSEServerParameters(BaseModel):
    """Parameters for the SSE server."""

    url: IsSSEURL


IsWSURL = Annotated[str, lambda v: v.startswith("ws://") or v.startswith("wss://")]


class WSServerParameters(BaseModel):
    """Parameters for the WS server."""

    url: IsWSURL


class MCPClient:
    def __init__(
        self,
        name: str,
        config: StdioServerParameters | SSEServerParameters | WSServerParameters,
    ):
        self.name = name

        self.client: Client | None = None
        try:
            cfg = StdioServerParameters.model_validate(config)
            self.transport = StdioTransport(
                command=cfg.command,
                args=cfg.args,
                env=cfg.env,
                cwd=str(cfg.cwd) if cfg.cwd else None,
            )
            return
        except ValidationError:
            pass
        try:
            cfg = SSEServerParameters.model_validate(config)
            self.transport = SSETransport(
                url=cfg.url,
            )
            return
        except (ValidationError, ValueError):
            pass
        try:
            cfg = WSServerParameters.model_validate(config)
            self.transport = WSTransport(
                url=cfg.url,
            )
        except ValidationError:
            raise ValueError("Invalid transport type")

    async def initialize(self) -> Client | None:
        """Initialize the server connection.

        Returns:
            The initialized client or None if initialization fails.
        """

        # We set up "done" as a future to signal when the client should shutdown.
        self.closed = asyncio.Event()
        self.done = asyncio.Event()
        # We wait for the client to be initialized before returning from initialize()
        client_initialized = asyncio.Event()

        async def task():
            assert self.transport is not None, "Transport is not initialized"
            try:
                async with Client(
                    self.transport,
                    log_handler=Logger(),
                    sampling_handler=sampling_handler,
                ) as client:
                    self.client = client
                    client_initialized.set()
                    await self.done.wait()
                self.closed.set()

            except Exception as e:
                log.error(f"Error initializing MCP server: {e}")
                client_initialized.set()

        self.task = asyncio.create_task(task())
        try:
            await asyncio.wait_for(client_initialized.wait(), timeout=5)
        except asyncio.TimeoutError:
            self.client = None
            log.error("Timeout while initializing MCP server", self.name)
        return self.client

    async def get_available_tools(self) -> list[MCPTool]:
        """List available tools from the server.

        Returns:
            A list of available tools.

        Raises:
            RuntimeError: If the server is not initialized.
        """
        if self.client is None:
            raise RuntimeError("Client is not initialized")
        try:
            tools_response = await self.client.list_tools()
        except McpError:
            return []

        return tools_response

    async def get_available_prompts(self) -> list[Prompt]:
        """List available prompts from the server.

        Returns:
            A list of available prompts.

        Raises:
            RuntimeError: If the server is not initialized.
        """
        if self.client is None:
            raise RuntimeError(f"Server {self.name} not initialized")

        try:
            prompts = await self.client.list_prompts()
        except McpError:
            return []

        for prompt in prompts:
            log.info(f"Loaded prompt {prompt.name} from {self.name}")

        return prompts

    async def call_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any],
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        """Execute a tool

        Args:
            tool_name: Name of the tool to execute.
            arguments: Tool arguments.

        Returns:
            Tool execution result.

        Raises:
            RuntimeError: If server is not initialized.
            Exception: If tool execution fails after all retries.
        """
        if not self.client:
            raise RuntimeError(f"Server {self.name} not initialized")

        try:
            result = await self.client.call_tool(tool_name, arguments)
            return result
        except Exception as e:
            log.error(f"Error executing tool: {e}.")
            return []

    async def call_prompt(
        self,
        prompt_name: str,
        arguments: dict[str, str],
    ) -> list[PromptMessage]:
        """Execute a prompt

        Args:
            prompt_name: Name of the prompt
            arguments: Prompt arguments.
        """

        if self.client is None:
            raise RuntimeError(f"Server {self.name} not initialized")
        try:
            result = await self.client.get_prompt(prompt_name, arguments)
            return result.messages
        except Exception as e:
            log.error(f"Error getting prompt: {e}.")
            return []

    async def teardown(self) -> None:
        if self.client is None:
            raise RuntimeError("Client is already closed")
        self.done.set()
        await self.closed.wait()
        self.client = None
        self.transport = None



import pytest

from oterm.tools.date_time import DateTimeTool, date_time
from oterm.tools.external import load_external_tools


@pytest.mark.asyncio
async def test_loading_custom_tool():
    # Test loading a callable from a well-defined module
    tools = load_external_tools(
        [
            {
                "tool": "oterm.tools.date_time:DateTimeTool",
                "callable": "oterm.tools.date_time:date_time",
            }
        ]
    )

    assert len(tools) == 1
    assert tools[0]["tool"] == DateTimeTool
    assert tools[0]["callable"] == date_time



import json
import os
from pathlib import Path

from dotenv import load_dotenv
from pydantic import BaseModel

from oterm.utils import get_default_data_dir

load_dotenv()


class EnvConfig(BaseModel):
    ENV: str = "development"
    OLLAMA_HOST: str = "127.0.0.1:11434"
    OLLAMA_URL: str = ""
    OTERM_VERIFY_SSL: bool = True
    OTERM_DATA_DIR: Path = get_default_data_dir()
    OPEN_WEATHER_MAP_API_KEY: str = ""


envConfig = EnvConfig.model_validate(os.environ)
if envConfig.OLLAMA_URL == "":
    envConfig.OLLAMA_URL = f"http://{envConfig.OLLAMA_HOST}"


class AppConfig:
    def __init__(self, path: Path | None = None):
        if path is None:
            path = envConfig.OTERM_DATA_DIR / "config.json"
        self._path = path
        self._data = {
            "theme": "textual-dark",
            "splash-screen": True,
        }
        try:
            with open(self._path) as f:
                saved = json.load(f)
                self._data = self._data | saved
        except FileNotFoundError:
            Path.mkdir(self._path.parent, parents=True, exist_ok=True)
            self.save()

    def set(self, key, value):
        self._data[key] = value
        self.save()

    def get(self, key, default=None):
        return self._data.get(key, default)

    def save(self):
        with open(self._path, "w") as f:
            json.dump(self._data, f)


# Expose AppConfig object for app to import
appConfig = AppConfig()



from collections.abc import Iterable
from pathlib import Path

from PIL import Image as PILImage
from textual.message import Message
from textual.widgets import DirectoryTree

IMG_MAX_SIZE = 80
IMAGE_EXTENSIONS = PILImage.registered_extensions()


class ImageAdded(Message):
    def __init__(self, path: Path, image: str) -> None:
        self.path = path
        self.image = image
        super().__init__()


class ImageDirectoryTree(DirectoryTree):
    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        return [
            path for path in paths if path.suffix in IMAGE_EXTENSIONS or path.is_dir()
        ]



import itertools

from oterm.types import ToolCall

available_tool_defs: dict[str, list[ToolCall]] = {}


def available_tool_calls() -> list[ToolCall]:
    return list(itertools.chain.from_iterable(available_tool_defs.values()))



import pytest
from mcp.types import Tool as MCPTool

from oterm.ollamaclient import OllamaLLM
from oterm.tools.mcp.client import MCPClient
from oterm.tools.mcp.tools import MCPToolCallable
from oterm.types import Tool


@pytest.mark.asyncio
async def test_mcp_tools(mcp_client: MCPClient, default_model):
    tools = await mcp_client.get_available_tools()
    for oracle in tools:
        assert MCPTool.model_validate(oracle)

    oracle = tools[0]
    oterm_tool = Tool(
        function=Tool.Function(
            name=oracle.name,
            description=oracle.description,
            parameters=Tool.Function.Parameters.model_validate(oracle.inputSchema),
        ),
    )

    mcpToolCallable = MCPToolCallable(oracle.name, "test_server", mcp_client)
    llm = OllamaLLM(
        model=default_model,
        tool_defs=[{"tool": oterm_tool, "callable": mcpToolCallable.call}],
    )

    res = await llm.completion("Ask the oracle what is the best client for Ollama.")
    assert (
        "oterm" in res or "orterm" in res
    )  # wtf is with orterm being the best client?



from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def add_id_to_messages(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                CREATE TABLE message_temp (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    chat_id INTEGER NOT NULL,
                    author TEXT NOT NULL,
                    text TEXT NOT NULL
                );
                INSERT INTO message_temp (chat_id, author, text) SELECT chat_id, author, text FROM message;
                DROP TABLE message;
                ALTER TABLE message_temp RENAME TO message;
                """
            )
        except aiosqlite.OperationalError:
            pass


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.5.1", [add_id_to_messages])
]



import rich.repr
from textual import Logger, LogGroup  # type: ignore

log_lines: list[tuple[LogGroup, str]] = []


@rich.repr.auto
class OtermLogger(Logger):
    @property
    def debug(self) -> "OtermLogger":
        """Logs debug messages."""
        return OtermLogger(self._log, LogGroup.DEBUG)

    @property
    def info(self) -> "OtermLogger":
        """Logs information."""
        res = OtermLogger(self._log, LogGroup.INFO)
        return res

    @property
    def warning(self) -> "OtermLogger":
        """Logs warnings."""
        return OtermLogger(self._log, LogGroup.WARNING)

    @property
    def error(self) -> "OtermLogger":
        """Logs errors."""
        return OtermLogger(self._log, LogGroup.ERROR)

    def __call__(self, *args: object, **kwargs) -> None:
        output = " ".join(str(arg) for arg in args)
        if kwargs:
            key_values = " ".join(f"{key}={value!r}" for key, value in kwargs.items())
            output = f"{output} {key_values}" if output else key_values
        log_lines.append((self._group, output))
        super().__call__(*args, **kwargs)


log = OtermLogger(None)



from mcp import Tool as MCPTool
from mcp.types import TextContent

from oterm.tools.mcp.client import MCPClient
from oterm.types import Tool


class MCPToolCallable:
    def __init__(self, name: str, server_name: str, client: MCPClient):
        self.name = name
        self.server_name = server_name
        self.client = client

    async def call(self, **kwargs) -> str:
        res = await self.client.call_tool(self.name, kwargs)
        text_content = [m.text for m in res if type(m) is TextContent]
        return "\n".join(text_content)


def mcp_tool_to_ollama_tool(mcp_tool: MCPTool) -> Tool:
    """Convert an MCP tool to an Ollama tool"""

    return Tool(
        function=Tool.Function(
            name=mcp_tool.name,
            description=mcp_tool.description,
            parameters=Tool.Function.Parameters.model_validate(mcp_tool.inputSchema),
        ),
    )



from oterm.types import Tool

ThinkTool = Tool(
    type="function",
    function=Tool.Function(
        name="think",
        description="Use the tool to think about something. It will not obtain new information or change the database, but just append the thought to the log. Use it when complex reasoning or some cache memory is needed.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={
                "thought": Tool.Function.Parameters.Property(
                    type="string", description="A thought to think about."
                ),
            },
            required=["thought"],
        ),
    ),
)


async def think(thought: str) -> str:
    return thought



from oterm.utils import int_to_semantic_version, semantic_version_to_int


def test_sqlite_user_version():
    version = "0.1.5"
    assert semantic_version_to_int(version) == 261
    assert int_to_semantic_version(261) == version

    version = "0.0.0"
    assert semantic_version_to_int(version) == 0
    assert int_to_semantic_version(0) == version

    version = "255.255.255"
    assert semantic_version_to_int(version) == 16777215
    assert int_to_semantic_version(16777215) == version



from mcp.client.session import LoggingFnT
from mcp.types import LoggingMessageNotificationParams

from oterm.log import log


# This is here to log the messages from the MCP server, when
# there is an upstream fix for https://github.com/modelcontextprotocol/python-sdk/issues/341
class Logger(LoggingFnT):
    async def __call__(self, params: LoggingMessageNotificationParams) -> None:
        if params.level == "error" or params.level == "critical":
            log.error(params.data)
        elif params.level == "warning":
            log.warning(params.data)
        elif params.level == "info":
            log.info(params.data)
        elif params.level == "debug":
            log.debug(params.data)



import asyncio
from importlib import metadata

import typer
from rich.pretty import pprint

from oterm.app.oterm import app
from oterm.config import envConfig
from oterm.store.store import Store

cli = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})


async def upgrade_db():
    await Store.get_store()


@cli.command()
def oterm(
    version: bool = typer.Option(None, "--version", "-v"),
    upgrade: bool = typer.Option(None, "--upgrade"),
    config: bool = typer.Option(None, "--config"),
    sqlite: bool = typer.Option(None, "--db"),
    data_dir: bool = typer.Option(None, "--data-dir"),
):
    if version:
        typer.echo(f"oterm v{metadata.version('oterm')}")
        exit(0)
    if upgrade:
        asyncio.run(upgrade_db())
        exit(0)
    if sqlite:
        typer.echo(envConfig.OTERM_DATA_DIR / "store.db")
        exit(0)
    if data_dir:
        typer.echo(envConfig.OTERM_DATA_DIR)
        exit(0)
    if config:
        typer.echo(pprint(envConfig))
        exit(0)
    app.run()


if __name__ == "__main__":
    cli()



import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.shell import ShellTool, shell_command


@pytest.mark.asyncio
async def test_shell(default_model):
    llm = OllamaLLM(
        model=default_model,
        tool_defs=[
            {"tool": ShellTool, "callable": shell_command},
        ],
    )
    res = await llm.completion(
        "What is the current directory? Use the shell tool available and execute the command."
    )
    assert "oterm" in res



from datetime import datetime

import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.date_time import DateTimeTool, date_time


@pytest.mark.asyncio
async def test_date_time(default_model):
    llm = OllamaLLM(
        model=default_model, tool_defs=[{"tool": DateTimeTool, "callable": date_time}]
    )
    res = await llm.completion(
        "What is the time in 24h format? Use the date_time tool to answer this question."
    )
    time = datetime.time(datetime.now())
    assert f"{time.hour:02}:{time.minute:02}" in res



from mcp import Tool as MCPTool

from oterm.config import appConfig
from oterm.log import log
from oterm.tools.mcp.client import MCPClient
from oterm.tools.mcp.prompts import MCPPromptCallable
from oterm.tools.mcp.tools import MCPToolCallable, mcp_tool_to_ollama_tool
from oterm.types import PromptCall, ToolCall

mcp_clients: list[MCPClient] = []


async def setup_mcp_servers() -> tuple[
    dict[str, list[ToolCall]], dict[str, list[PromptCall]]
]:
    mcp_servers = appConfig.get("mcpServers")
    tool_calls: dict[str, list[ToolCall]] = {}
    prompt_calls: dict[str, list[PromptCall]] = {}

    if mcp_servers:
        for server, config in mcp_servers.items():
            client = MCPClient(server, config)
            await client.initialize()
            if not client.client:
                continue
            mcp_clients.append(client)

            log.info(f"Initialized MCP server {server}")

            mcp_tools: list[MCPTool] = await client.get_available_tools()
            mcp_prompts = await client.get_available_prompts()

            if mcp_tools:
                tool_calls[server] = []
            for mcp_tool in mcp_tools:
                tool = mcp_tool_to_ollama_tool(mcp_tool)
                mcpToolCallable = MCPToolCallable(mcp_tool.name, server, client)
                tool_calls[server].append(
                    {"tool": tool, "callable": mcpToolCallable.call}
                )
                log.info(f"Loaded MCP tool {mcp_tool.name} from {server}")

            if mcp_prompts:
                prompt_calls[server] = []

            for prompt in mcp_prompts:
                mcpPromptCallable = MCPPromptCallable(prompt.name, server, client)
                prompt_calls[server].append(
                    {"prompt": prompt, "callable": mcpPromptCallable.call}
                )
                log.info(f"Loaded MCP prompt {prompt.name} from {server}")

    return tool_calls, prompt_calls


async def teardown_mcp_servers():
    log.info("Tearing down MCP servers")
    # Important to tear down in reverse order
    mcp_clients.reverse()
    for client in mcp_clients:
        await client.teardown()



from dataclasses import dataclass
from typing import cast

from textual import events, on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Button, Input, TextArea

from oterm.app.image_browser import ImageSelect
from oterm.app.widgets.image import ImageAdded


class PostableTextArea(TextArea):
    """
    A text area that submits on Enter.
    """

    BINDINGS = TextArea.BINDINGS + [
        Binding(
            key="enter",
            action="submit",
            description="submit",
            show=True,
            key_display=None,
            priority=True,
        ),
        Binding(
            key="shift+enter",
            action="newline",
            description="newline",
            show=True,
            key_display=None,
            priority=True,
            id="newline",
        ),
    ]

    @dataclass
    class Submitted(Message):
        input: "PostableTextArea"
        value: str

        @property
        def control(self) -> "PostableTextArea":
            return self.input

    def action_submit(self) -> None:
        self.post_message(PostableTextArea.Submitted(self, self.text))

    def action_newline(self) -> None:
        cur = self.cursor_location
        lines = self.text.split("\n")
        lines[cur[0]] = lines[cur[0]][: cur[1]] + "\n" + lines[cur[0]][cur[1] :]
        self.text = "\n".join(lines)
        self.cursor_location = (cur[0] + 1, 0)


class PastableInput(Input):
    BINDINGS = Input.BINDINGS + [
        Binding(
            key="ctrl+m",
            action="toggle_multiline",
            description="multiline",
            show=True,
            key_display=None,
            priority=True,
            id="toggle.multiline",
        ),
    ]

    def action_toggle_multiline(self) -> None:
        input = cast(FlexibleInput, self.parent.parent)  # type: ignore
        input.text = self.value
        input.toggle_multiline()

    def _on_paste(self, event: events.Paste) -> None:
        if event.text:
            self.insert_text_at_cursor(event.text)
            if len(event.text.splitlines()) > 1:
                input = cast(FlexibleInput, self.parent.parent)  # type: ignore
                input.text = self.value
                input.toggle_multiline()
        event.stop()
        event.prevent_default()


class FlexibleInput(Widget):
    is_multiline = reactive(False)
    text = reactive("")

    BINDINGS = [
        Binding("ctrl+i", "add_image", "add image", id="add.image"),
    ]

    @dataclass
    class Submitted(Message):
        input: "FlexibleInput"
        value: str

        @property
        def control(self) -> "FlexibleInput":
            return self.input

    def __init__(self, text, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.text = text

    def on_mount(self) -> None:
        input = self.query_one("#promptInput", PastableInput)
        textarea = self.query_one("#promptArea", TextArea)
        textarea.show_line_numbers = False
        input.focus()

    def clear(self) -> None:
        self.text = ""
        self.query_one("#promptInput", PastableInput).value = ""
        self.query_one("#promptArea", TextArea).text = ""

    def focus(self, scroll_visible=True) -> "FlexibleInput":
        if self.is_multiline:
            self.query_one("#promptArea", TextArea).focus()
        else:
            self.query_one("#promptInput", PastableInput).focus()
        return self

    def toggle_multiline(self):
        self.is_multiline = not self.is_multiline
        input = self.query_one("#promptInput", PastableInput)
        textarea = self.query_one("#promptArea", TextArea)
        if self.is_multiline:
            textarea.text = self.text
            self.add_class("multiline")
            self.remove_class("singleline")
            line_count = textarea.document.line_count
            textarea.cursor_location = (
                line_count - 1,
                len(textarea.document.get_line(line_count - 1)),
            )
        else:
            input.value = self.text
            self.add_class("singleline")
            self.remove_class("multiline")
        self.focus()

    def watch_text(self):
        try:
            if len(self.text.splitlines()) > 1:
                self.query_one("#toggle-multiline", Button).disabled = True
            else:
                self.query_one("#toggle-multiline", Button).disabled = False

            input = self.query_one("#promptInput", PastableInput)
            textarea = self.query_one("#promptArea", TextArea)
            if self.is_multiline:
                if textarea.text != self.text:
                    textarea.text = self.text
            else:
                if input.value != self.text:
                    input.value = self.text
        except NoMatches:
            pass

    def action_add_image(self) -> None:
        async def on_image_selected(image) -> None:
            if image is None:
                return
            path, b64 = image
            self.post_message(ImageAdded(path, b64))

        screen = ImageSelect()
        self.app.push_screen(screen, on_image_selected)

    @on(PastableInput.Submitted, "#promptInput")
    def on_input_submitted(self, event: PastableInput.Submitted):
        self.post_message(self.Submitted(self, event.input.value))
        event.stop()
        event.prevent_default()

    @on(PostableTextArea.Submitted, "#promptArea")
    def on_textarea_submitted(self, event: PostableTextArea.Submitted):
        self.post_message(self.Submitted(self, event.input.text))
        event.stop()
        event.prevent_default()

    @on(Button.Pressed, "#toggle-multiline")
    def on_toggle_multiline_pressed(self):
        self.toggle_multiline()

    @on(PastableInput.Changed, "#promptInput")
    def on_input_changed(self, event: PastableInput.Changed):
        self.text = event.input.value

    @on(TextArea.Changed, "#promptArea")
    def on_area_changed(self, event: TextArea.Changed):
        lines = [
            event.text_area.document.get_line(line)
            for line in range(event.text_area.document.line_count)
        ]

        self.text = "\n".join(lines)

    @on(Button.Pressed, "#post")
    async def on_post(self):
        self.post_message(self.Submitted(self, self.text))

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield PastableInput(
                id="promptInput",
                placeholder="Your message…",
            )
            yield PostableTextArea(id="promptArea")
            with Horizontal(id="button-container"):
                yield Button("post", id="post", variant="primary")
                yield Button(
                    "↕", id="toggle-multiline", classes="icon", variant="success"
                )



from fastmcp.client.transports import SSETransport, StdioTransport, WSTransport

from oterm.tools.mcp.client import MCPClient


def test_stdio_transport(mcp_server_config):
    """
    Test the MCP client with a StdioServerParameters.
    """

    client = MCPClient("test_stdio", mcp_server_config["stdio"])
    assert isinstance(client.transport, StdioTransport)


def test_sse_transport(mcp_server_config):
    client = MCPClient("test_sse", mcp_server_config["sse"])
    assert isinstance(client.transport, SSETransport)


def test_ws_transport(mcp_server_config):
    client = MCPClient("test_ws", mcp_server_config["ws"])
    assert isinstance(client.transport, WSTransport)



from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def images(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                ALTER TABLE message ADD COLUMN images TEXT DEFAULT "[]";
                """
            )
        except aiosqlite.OperationalError:
            pass

        await connection.commit()


async def orphan_messages(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                DELETE FROM message WHERE chat_id NOT IN (SELECT id FROM chat);
                """
            )
        except aiosqlite.OperationalError:
            pass

        await connection.commit()


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.7.0", [images, orphan_messages]),
]



from collections.abc import Iterable

from textual import on, work
from textual.app import App, ComposeResult, SystemCommand
from textual.binding import Binding
from textual.screen import Screen
from textual.widgets import Footer, Header, TabbedContent, TabPane

from oterm.app.chat_edit import ChatEdit
from oterm.app.chat_export import ChatExport, slugify
from oterm.app.pull_model import PullModel
from oterm.app.splash import splash
from oterm.app.widgets.chat import ChatContainer
from oterm.config import appConfig
from oterm.store.store import Store
from oterm.tools.external import load_external_tools
from oterm.tools.mcp.setup import setup_mcp_servers, teardown_mcp_servers
from oterm.types import ChatModel, ExternalToolDefinition
from oterm.utils import check_ollama, is_up_to_date


class OTerm(App):
    TITLE = "oterm"
    SUB_TITLE = "the TUI Ollama client."
    CSS_PATH = "oterm.tcss"
    BINDINGS = [
        Binding("ctrl+tab", "cycle_chat(+1)", "next chat", id="next.chat"),
        Binding("ctrl+shift+tab", "cycle_chat(-1)", "prev chat", id="prev.chat"),
        Binding("ctrl+backspace", "delete_chat", "delete chat", id="delete.chat"),
        Binding("ctrl+n", "new_chat", "new chat", id="new.chat"),
        Binding("ctrl+l", "show_logs", "show logs", id="show.logs"),
        Binding("ctrl+q", "quit", "quit", id="quit"),
    ]

    def get_system_commands(self, screen: Screen) -> Iterable[SystemCommand]:
        yield from super().get_system_commands(screen)
        yield SystemCommand("New chat", "Creates a new chat", self.action_new_chat)
        yield SystemCommand(
            "Edit chat parameters",
            "Allows to redefine model parameters and system prompt",
            self.action_edit_chat,
        )
        yield SystemCommand(
            "Rename chat", "Renames the current chat", self.action_rename_chat
        )
        yield SystemCommand(
            "Clear chat", "Clears the current chat", self.action_clear_chat
        )
        yield SystemCommand(
            "Delete chat", "Deletes the current chat", self.action_delete_chat
        )
        yield SystemCommand(
            "Export chat",
            "Exports the current chat as Markdown (in the current working directory)",
            self.action_export_chat,
        )
        yield SystemCommand(
            "Regenerate last Ollama message",
            "Regenerates the last Ollama message (setting a random seed for the message)",
            self.action_regenerate_last_message,
        )
        yield SystemCommand(
            "Use MCP prompt",
            "Create and copy to clipboard an MCP prompt.",
            self.action_mcp_prompt,
        )
        yield SystemCommand(
            "Pull model",
            "Pulls (or updates) the model from the Ollama server",
            self.action_pull_model,
        )
        yield SystemCommand(
            "Show logs", "Shows the logs of the app", self.action_show_logs
        )

    async def action_quit(self) -> None:
        self.log("Quitting...")
        await teardown_mcp_servers()
        return self.exit()

    async def action_cycle_chat(self, change: int) -> None:
        tabs = self.query_one(TabbedContent)
        store = await Store.get_store()
        saved_chats = await store.get_chats()
        if tabs.active_pane is None:
            return
        active_id = int(str(tabs.active_pane.id).split("-")[1])
        for chat_model in saved_chats:
            if chat_model.id == active_id:
                next_index = (saved_chats.index(chat_model) + change) % len(saved_chats)
                next_id = saved_chats[next_index].id
                if next_id is not None:  # Ensure we have a valid ID
                    tabs.active = f"chat-{next_id}"
                break

    @work
    async def action_new_chat(self) -> None:
        store = await Store.get_store()
        model_info: str | None = await self.push_screen_wait(ChatEdit())
        if not model_info:
            return

        chat_model = ChatModel.model_validate_json(model_info)
        tabs = self.query_one(TabbedContent)
        tab_count = tabs.tab_count

        name = f"chat #{tab_count + 1} - {chat_model.model}"
        chat_model.name = name

        id = await store.save_chat(chat_model)
        chat_model.id = id

        pane = TabPane(name, id=f"chat-{id}")
        pane.compose_add_child(
            ChatContainer(
                chat_model=chat_model,
                messages=[],
            )
        )
        await tabs.add_pane(pane)
        tabs.active = f"chat-{id}"

    async def action_edit_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        chat.action_edit_chat()

    async def action_rename_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        chat.action_rename_chat()

    async def action_clear_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        await chat.action_clear_chat()

    async def action_delete_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        store = await Store.get_store()

        if chat.chat_model.id is not None:
            await store.delete_chat(chat.chat_model.id)
            await tabs.remove_pane(tabs.active)
            self.notify(f"Deleted {chat.chat_model.name}", severity="information")

    async def action_export_chat(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)

        if chat.chat_model.id is not None:
            screen = ChatExport(
                chat_id=chat.chat_model.id,
                file_name=f"{slugify(chat.chat_model.name)}.md",
            )
            self.push_screen(screen)

    async def action_regenerate_last_message(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        await chat.action_regenerate_llm_message()

    async def action_mcp_prompt(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            return
        chat = tabs.active_pane.query_one(ChatContainer)
        chat.action_mcp_prompt()

    async def action_pull_model(self) -> None:
        tabs = self.query_one(TabbedContent)
        if tabs.active_pane is None:
            screen = PullModel("")
        else:
            chat = tabs.active_pane.query_one(ChatContainer)
            screen = PullModel(chat.ollama.model)
        self.push_screen(screen)

    async def action_show_logs(self) -> None:
        from oterm.app.log_viewer import LogViewer

        screen = LogViewer()
        self.push_screen(screen)

    async def load_mcp(self):
        from oterm.tools import available_tool_defs
        from oterm.tools.mcp.prompts import available_prompt_defs

        external_tool_defs: list[ExternalToolDefinition] = appConfig.get("tools", [])  # type: ignore
        external_tools = list(load_external_tools(external_tool_defs))
        available_tool_defs["external"] = external_tools
        mcp_tool_calls, mcp_prompt_calls = await setup_mcp_servers()
        available_tool_defs.update(mcp_tool_calls)
        available_prompt_defs.update(mcp_prompt_calls)

    @work(exclusive=True)
    async def perform_checks(self) -> None:
        await check_ollama()
        up_to_date, _, latest = await is_up_to_date()
        if not up_to_date:
            self.notify(
                f"[b]oterm[/b] version [i]{latest}[/i] is available, please update.",
                severity="warning",
            )

    async def on_mount(self) -> None:
        store = await Store.get_store()
        theme = appConfig.get("theme")
        if theme:
            if theme == "dark":
                self.theme = "textual-dark"
            elif theme == "light":
                self.theme = "textual-light"
            else:
                self.theme = theme
        self.dark = appConfig.get("theme") == "dark"
        self.watch(self.app, "theme", self.on_theme_change, init=False)

        saved_chats = await store.get_chats()
        # Apply any remap of key bindings.
        keymap = appConfig.get("keymap")
        if keymap:
            self.set_keymap(keymap)

        await self.load_mcp()

        async def on_splash_done(message) -> None:
            if not saved_chats:
                # Pyright suggests awaiting here which has bitten me twice
                # so I'm ignoring it
                self.action_new_chat()  # type: ignore
            else:
                tabs = self.query_one(TabbedContent)
                for chat_model in saved_chats:
                    # Only process chats with a valid ID
                    if chat_model.id is not None:
                        messages = await store.get_messages(chat_model.id)
                        container = ChatContainer(
                            chat_model=chat_model,
                            messages=messages,
                        )
                        pane = TabPane(
                            chat_model.name, container, id=f"chat-{chat_model.id}"
                        )
                        tabs.add_pane(pane)
            self.perform_checks()

        if appConfig.get("splash-screen"):
            self.push_screen(splash, callback=on_splash_done)
        else:
            await on_splash_done("")

    def on_theme_change(self, old_value: str, new_value: str) -> None:
        if appConfig.get("theme") != new_value:
            appConfig.set("theme", new_value)

    @work
    @on(TabbedContent.TabActivated)
    async def on_tab_activated(self, event: TabbedContent.TabActivated) -> None:
        container = event.pane.query_one(ChatContainer)
        await container.load_messages()

    def compose(self) -> ComposeResult:
        yield Header()
        yield TabbedContent(id="tabs")
        yield Footer()


app = OTerm()



from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def update_format(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                UPDATE chat SET format = '' WHERE format is NULL;
                """
            )
        except aiosqlite.OperationalError:
            pass


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.2.4", [update_format])
]



from collections.abc import Awaitable, Callable
from pathlib import Path

import aiosqlite


async def keep_alive(db_path: Path) -> None:
    async with aiosqlite.connect(db_path) as connection:
        try:
            await connection.executescript(
                """
                ALTER TABLE chat ADD COLUMN keep_alive INTEGER DEFAULT 5;
                """
            )
        except aiosqlite.OperationalError:
            pass


upgrades: list[tuple[str, list[Callable[[Path], Awaitable[None]]]]] = [
    ("0.2.8", [keep_alive])
]



import inspect
import json
from collections.abc import Awaitable, Callable

from mcp.types import Prompt
from ollama import Message
from textual import on
from textual.app import ComposeResult, RenderResult
from textual.containers import (
    Container,
    Horizontal,
    Vertical,
    VerticalScroll,
)
from textual.screen import ModalScreen
from textual.widget import Widget
from textual.widgets import Button, Input, Label, OptionList, TextArea
from textual.widgets.option_list import Option

from oterm.tools.mcp.prompts import (
    available_prompt_calls,
    available_prompt_defs,
    mcp_prompt_to_ollama_messages,
)
from oterm.utils import debounce


class PromptOptionWidget(Widget):
    def __init__(self, server: str, prompt: Prompt) -> None:
        super().__init__()
        self.prompt = prompt
        self.server = server

    def render(self) -> RenderResult:
        return f"[b]{self.server} - {self.prompt.name}[/b]\n[i]{self.prompt.description}[/i]"


class PromptFormWidget(Widget):
    prompt: Prompt
    callable: Callable | Awaitable
    messages: list[Message] = []

    @on(Input.Changed)
    @debounce(1.0)
    async def on_text_area_change(self) -> None:
        is_valid = True
        params = {}
        for arg in self.prompt.arguments or []:
            params[arg.name] = self.query_one(f"#arg-{arg.name}", Input).value
            if arg.required and not params[arg.name]:
                is_valid = False
        prompt_result_widget = self.query_one("#prompt-result", TextArea)
        if inspect.iscoroutinefunction(self.callable):
            messages = await self.callable(**params)
        else:
            messages = self.callable(**params)  # type: ignore
        self.messages = messages = mcp_prompt_to_ollama_messages(messages)
        prompt_result_widget.text = "\n".join(
            [f"{m.role}: {m.content}" for m in messages]
        )
        submit_button = self.screen.query_one("#submit", Button)
        submit_button.disabled = not is_valid

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="prompt-form-container"):
            for arg in self.prompt.arguments or []:
                yield Label(
                    f"{arg.name}{arg.required and ' (required)' or ''}", classes="title"
                )
                yield Input(id=f"arg-{arg.name}", tooltip=arg.description)
            yield Label("Messages:", classes="subtitle")
            yield TextArea(id="prompt-result", read_only=True)


class MCPPrompt(ModalScreen[str]):
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
        ("enter", "submit", "Submit"),
    ]

    def action_cancel(self) -> None:
        self.dismiss()

    async def on_mount(self) -> None:
        option_list = self.query_one("#mcp-prompt-select", OptionList)
        option_list.clear_options()
        for server in available_prompt_defs.keys():
            for prompt_call in available_prompt_defs[server]:
                option_list.add_option(
                    option=self.prompt_option(server, prompt_call["prompt"])
                )

    @staticmethod
    def prompt_option(server: str, prompt: Prompt) -> Option:
        return Option(
            prompt=PromptOptionWidget(server, prompt).render(), id=prompt.name
        )

    async def on_option_list_option_highlighted(
        self, option: OptionList.OptionHighlighted
    ) -> None:
        prompt = None
        prompt_call = None
        for prompt_call in available_prompt_calls():
            prompt = prompt_call["prompt"]
            if prompt.name == option.option.id:
                break
        if prompt is None or prompt_call is None:
            return

        form_container = self.query_one("#prompt-form-container", Vertical)
        form_container = self.query_one("#prompt-form-container", Vertical)
        form_container.remove_children()
        widget = PromptFormWidget(classes="prompt-form")
        widget.prompt = prompt
        widget.callable = prompt_call["callable"]
        form_container.mount(widget)

        has_required_args = any(arg.required for arg in prompt.arguments or [])
        submit_button = self.query_one("#submit", Button)
        submit_button.disabled = has_required_args
        if not has_required_args:
            await widget.on_text_area_change()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.name == "submit":
            form = self.query_one(".prompt-form", PromptFormWidget)
            jsn = json.dumps([m.model_dump() for m in form.messages])
            self.dismiss(jsn)
        else:
            self.dismiss()

    def compose(self) -> ComposeResult:
        with Container(classes="screen-container full-height"):
            with Horizontal():
                with Vertical():
                    yield Label("Available MCP prompts", classes="title")
                    yield OptionList(id="mcp-prompt-select")

                with Vertical():
                    yield Label("Customize prompt:", classes="title")
                    yield Vertical(id="prompt-form-container")
            with Horizontal(classes="button-container"):
                yield Button(
                    "Submit",
                    id="submit",
                    name="submit",
                    variant="primary",
                    disabled=True,
                )
                yield Button("Cancel", name="cancel")



from collections.abc import Awaitable, Callable, Sequence
from importlib import import_module

from ollama import Tool

from oterm.log import log
from oterm.types import ExternalToolDefinition, ToolCall


def load_external_tools(
    external_tools: Sequence[ExternalToolDefinition],
) -> Sequence[ToolCall]:
    tools = []
    for tool_def in external_tools:
        tool_path = tool_def["tool"]

        try:
            module, tool = tool_path.split(":")
            module = import_module(module)
            tool = getattr(module, tool)
            if not isinstance(tool, Tool):
                raise Exception(f"Expected Tool, got {type(tool)}")
        except ModuleNotFoundError as e:
            log.error(f"Error loading tool {tool_path}: {e}")
            continue

        callable_path = tool_def["callable"]
        try:
            module, function = callable_path.split(":")
            module = import_module(module)
            callable = getattr(module, function)
            if not isinstance(callable, Callable | Awaitable):
                raise Exception(f"Expected Callable, got {type(callable)}")
        except ModuleNotFoundError as e:
            log.error(f"Error loading callable {callable_path}: {e}")
            continue
        log.info(f"Loaded tool {tool.function.name} from {tool_path}")  # type: ignore
        tools.append({"tool": tool, "callable": callable})

    return tools



from oterm.store.upgrades.v0_1_6 import upgrades as v0_1_6_upgrades
from oterm.store.upgrades.v0_1_11 import upgrades as v_0_1_11_upgrades
from oterm.store.upgrades.v0_2_0 import upgrades as v0_2_0_upgrades
from oterm.store.upgrades.v0_2_4 import upgrades as v0_2_4_upgrades
from oterm.store.upgrades.v0_2_8 import upgrades as v0_2_8_upgrades
from oterm.store.upgrades.v0_3_0 import upgrades as v0_3_0_upgrades
from oterm.store.upgrades.v0_4_0 import upgrades as v0_4_0_upgrades
from oterm.store.upgrades.v0_5_1 import upgrades as v0_5_1_upgrades
from oterm.store.upgrades.v0_6_0 import upgrades as v0_6_0_upgrades
from oterm.store.upgrades.v0_7_0 import upgrades as v0_7_0_upgrades
from oterm.store.upgrades.v0_9_0 import upgrades as v0_9_0_upgrades
from oterm.store.upgrades.v0_12_0 import upgrades as v0_12_0_upgrades
from oterm.store.upgrades.v0_13_1 import upgrades as v0_13_1_upgrades

upgrades = (
    v0_1_6_upgrades
    + v_0_1_11_upgrades
    + v0_2_0_upgrades
    + v0_2_4_upgrades
    + v0_2_8_upgrades
    + v0_3_0_upgrades
    + v0_4_0_upgrades
    + v0_5_1_upgrades
    + v0_6_0_upgrades
    + v0_7_0_upgrades
    + v0_9_0_upgrades
    + v0_12_0_upgrades
    + v0_13_1_upgrades
)



from base64 import b64encode
from io import BytesIO
from pathlib import Path

import PIL.Image as PILImage
from PIL import UnidentifiedImageError
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import DirectoryTree, Input, Label
from textual_image.widget import Image

from oterm.app.widgets.image import IMAGE_EXTENSIONS, ImageDirectoryTree


class ImageSelect(ModalScreen[tuple[Path, str]]):
    BINDINGS = [
        ("escape", "cancel", "Cancel"),
    ]

    def action_cancel(self) -> None:
        self.dismiss()

    async def on_mount(self) -> None:
        dt = self.query_one(ImageDirectoryTree)
        dt.show_guides = False
        dt.focus()

    @on(DirectoryTree.FileSelected)
    async def on_image_selected(self, ev: DirectoryTree.FileSelected) -> None:
        try:
            buffer = BytesIO()
            image = PILImage.open(ev.path)
            if image.mode != "RGB":
                image = image.convert("RGB")
            image.save(buffer, format="JPEG")
            b64 = b64encode(buffer.getvalue()).decode("utf-8")
            self.dismiss((ev.path, b64))
        except UnidentifiedImageError:
            self.dismiss()

    @on(DirectoryTree.NodeHighlighted)
    async def on_image_highlighted(self, ev: DirectoryTree.NodeHighlighted) -> None:
        path = ev.node.data.path  # type: ignore
        image_widget = self.query_one(Image)
        if path.suffix in IMAGE_EXTENSIONS:
            try:
                image_widget.image = PILImage.open(path.as_posix())
            except UnidentifiedImageError:
                image_widget.image = None
        else:
            image_widget.image = None

    @on(Input.Changed)
    async def on_root_changed(self, ev: Input.Changed) -> None:
        dt = self.query_one(ImageDirectoryTree)
        path = Path(ev.value)
        if not path.exists() or not path.is_dir():
            return
        dt.path = path

    def compose(self) -> ComposeResult:
        with Container(
            id="image-select-container", classes="screen-container full-height"
        ):
            with Horizontal():
                with Vertical(id="image-directory-tree"):
                    yield Label("Select an image:", classes="title")
                    yield Label("Root:")
                    yield Input(Path("./").resolve().as_posix())
                    yield ImageDirectoryTree("./")
                with Container(id="image-preview"):
                    yield Image(id="image")



import pytest

from oterm.ollamaclient import OllamaLLM
from oterm.tools.mcp.client import MCPClient
from oterm.tools.mcp.tools import MCPToolCallable
from oterm.types import Tool


@pytest.mark.asyncio
async def test_mcp_sampling(mcp_client: MCPClient, default_model):
    """
    Test the sampling capbilities of oterm.
    Here we go full circle and use the MCP client to call the server
    to call the client again with a sampling request.
    """

    await mcp_client.initialize()

    tools = await mcp_client.get_available_tools()
    puzzle_solver = tools[1]
    oterm_tool = Tool(
        function=Tool.Function(
            name=puzzle_solver.name,
            description=puzzle_solver.description,
            parameters=Tool.Function.Parameters.model_validate(
                puzzle_solver.inputSchema
            ),
        ),
    )
    mcpToolCallable = MCPToolCallable(puzzle_solver.name, "test_server", mcp_client)
    llm = OllamaLLM(
        model=default_model,
        tool_defs=[{"tool": oterm_tool, "callable": mcpToolCallable.call}],
    )

    res = await llm.completion(
        """
        Solve the following puzzle by calling the puzzle solver tool.
        Jack is looking at Anne. Anne is looking at George.
        Jack is married, George is not, and we don't know if Anne is married.
        Is a married person looking at an unmarried person?
        Just answer yes or no."""
    )
    assert "no" in res.lower() or "yes" in res.lower()



from ollama import Tool

OracleTool = Tool(
    type="function",
    function=Tool.Function(
        name="oracle",
        description="Function to return the Oracle's answer to any question.",
        parameters=Tool.Function.Parameters(
            type="object",
            properties={
                "question": Tool.Function.Parameters.Property(
                    type="str", description="The question to ask."
                ),
            },
            required=["question"],
        ),
    ),
)


def oracle(question: str):
    return "oterm"



import asyncio
import sys
from collections.abc import Callable
from functools import wraps
from importlib import metadata
from pathlib import Path

import httpx
from packaging.version import Version, parse

from oterm.types import ParsedResponse


def debounce(wait: float) -> Callable:
    """
    A decorator to debounce a function, ensuring it is called only after a specified delay
    and always executes after the last call.

    Args:
        wait (float): The debounce delay in seconds.

    Returns:
        Callable: The decorated function.
    """

    def decorator(func: Callable) -> Callable:
        last_call = None
        task = None

        @wraps(func)
        async def debounced(*args, **kwargs):
            nonlocal last_call, task
            last_call = asyncio.get_event_loop().time()

            if task:
                task.cancel()

            async def call_func():
                await asyncio.sleep(wait)
                if asyncio.get_event_loop().time() - last_call >= wait:  # type: ignore
                    await func(*args, **kwargs)

            task = asyncio.create_task(call_func())

        return debounced

    return decorator


def parse_response(input_text: str) -> ParsedResponse:
    """
    Parse a response from the chatbot.
    """

    thought = ""
    response = input_text
    formatted_output = input_text

    # If the response contains a think tag, split the response into the thought process and the actual response
    thought_end = input_text.find("</think>")
    if input_text.startswith("<think>") and thought_end != -1:
        thought = input_text[7:thought_end].lstrip("\n").rstrip("\n").strip()
        response = input_text[thought_end + 8 :].lstrip("\n").rstrip("\n")
        # transform the think tag into a markdown blockquote (for clarity)
        if thought.strip():
            thought = "\n".join([f"> {line}" for line in thought.split("\n")])
            formatted_output = (
                "> ### <thought\\>\n" + thought + "\n> ### </thought\\>\n" + response
            )

    return ParsedResponse(
        thought=thought, response=response, formatted_output=formatted_output
    )


def get_default_data_dir() -> Path:
    """
    Get the user data directory for the current system platform.

    Linux: ~/.local/share/oterm
    macOS: ~/Library/Application Support/oterm
    Windows: C:/Users/<USER>/AppData/Roaming/oterm

    :return: User Data Path
    :rtype: Path
    """
    home = Path.home()

    system_paths = {
        "win32": home / "AppData/Roaming/oterm",
        "linux": home / ".local/share/oterm",
        "darwin": home / "Library/Application Support/oterm",
    }

    data_path = system_paths[sys.platform]
    return data_path


def semantic_version_to_int(version: str) -> int:
    """
    Convert a semantic version string to an integer.

    :param version: Semantic version string
    :type version: str
    :return: Integer representation of semantic version
    :rtype: int
    """
    major, minor, patch = version.split(".")
    major = int(major) << 16
    minor = int(minor) << 8
    patch = int(patch)
    return major + minor + patch


def int_to_semantic_version(version: int) -> str:
    """
    Convert an integer to a semantic version string.

    :param version: Integer representation of semantic version
    :type version: int
    :return: Semantic version string
    :rtype: str
    """
    major = version >> 16
    minor = (version >> 8) & 255
    patch = version & 255
    return f"{major}.{minor}.{patch}"


async def is_up_to_date() -> tuple[bool, Version, Version]:
    """
    Checks whether oterm is current.

    :return: A tuple containing a boolean indicating whether oterm is current, the running version and the latest version
    :rtype: tuple[bool, Version, Version]
    """

    async with httpx.AsyncClient() as client:
        running_version = parse(metadata.version("oterm"))
        try:
            response = await client.get("https://pypi.org/pypi/oterm/json")
            data = response.json()
            pypi_version = parse(data["info"]["version"])
        except Exception:
            # If no network connection, do not raise alarms.
            pypi_version = running_version
    return running_version >= pypi_version, running_version, pypi_version


async def check_ollama() -> bool:
    """
    Check if the Ollama server is up and running
    """
    from oterm.config import envConfig

    up = False
    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(envConfig.OLLAMA_URL)
            up = response.status_code == 200
    except httpx.HTTPError:
        up = False
    finally:
        if not up:
            from oterm.app.oterm import app

            app.notify(
                f"The Ollama server is not reachable at {envConfig.OLLAMA_URL}, please check your connection or set the OLLAMA_URL environment variable. oterm will now quit.",
                severity="error",
                timeout=10,
            )

            async def quit():
                await asyncio.sleep(10.0)
                try:
                    from oterm.tools.mcp.setup import teardown_mcp_servers

                    await teardown_mcp_servers()
                    exit()

                except Exception:
                    pass

            asyncio.create_task(quit())
    return up



from mcp import SamplingMessage
from mcp.server.fastmcp import Context, FastMCP
from mcp.server.fastmcp.prompts.base import AssistantMessage, Message, UserMessage
from mcp.types import ModelHint, ModelPreferences, TextContent

mcp = FastMCP("TestServer", port=8080)


@mcp.resource("config://app")
def get_config() -> str:
    return "Oracle MCP server"


@mcp.tool(name="oracle", description="Ask the oracle a question.")
async def oracle(query: str, ctx: Context) -> str:
    return "Oracle says: oterm"


@mcp.tool(name="puzzle_solver", description="Solves a puzzle by asking an advanced AI.")
async def puzzle_solver(puzzle_description: str, ctx: Context) -> str:
    """
    This tool is included to make a sampling request to the server.
    It takes a puzzle description and returns the answer.
    """
    session = ctx.session
    sampling_response = await session.create_message(
        messages=[
            SamplingMessage(
                role="user",
                content=TextContent(
                    text=f"Please solve this puzzle: {puzzle_description}", type="text"
                ),
            )
        ],
        model_preferences=ModelPreferences(
            hints=[ModelHint(name="mistral")],
        ),
        max_tokens=100,
    )
    return sampling_response.content.text  # type: ignore


@mcp.prompt(name="oracle_prompt", description="Prompt to ask the oracle a question.")
async def oracle_prompt(question: str) -> str:
    return f"Oracle: {question}"


@mcp.prompt(name="debug_error", description="Prompt to debug an error.")
async def debug_error(error: str, language: str = "python") -> list[Message]:
    return [
        UserMessage(f"I'm seeing this {language} error: {error}"),
        AssistantMessage("I'll help debug that. What have you tried so far?"),
    ]


if __name__ == "__main__":
    mcp.run(transport="stdio")
